'''
   MFEM example 3

   See c++ version in the MFEM library for more detail 
'''
import mfem.ser as mfem
from mfem.ser import intArray
import os
from os.path import expanduser, join
import numpy as np
from numpy import sin, array


def run(order=1,
        static_cond=False,
        freq=1,
        meshfile='',
        visualization=False,
        device='cpu',
        numba=False,
        pa=False):

    kappa = np.pi*freq

    device = mfem.Device(device)
    device.Print()

    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    if numba:
        #
        #  mfem.jit supports two interface. interface=simple allows for
        #  returning a numpy array.
        #
        @mfem.jit.vector(vdim=3, interface="simple")
        def E_exact(x):
            return np.array((sin(kappa*x[1]),
                             sin(kappa*x[2]),
                             sin(kappa*x[0])))

        @mfem.jit.vector(vdim=3, interface="c++")
        def f_exact(x, out):
            out[0] = (1 + kappa**2)*sin(kappa * x[1])
            out[1] = (1 + kappa**2)*sin(kappa * x[2])
            out[2] = (1 + kappa**2)*sin(kappa * x[0])
    else:
        class cE_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                return (sin(kappa * x[1]),
                        sin(kappa * x[2]),
                        sin(kappa * x[0]))
        E_exact = cE_exact()

        class cf_exact(mfem.VectorPyCoefficient):
            def __init__(self):
                mfem.VectorPyCoefficient.__init__(self, sdim)

            def EvalValue(self, x):
                return ((1 + kappa**2)*sin(kappa * x[1]),
                        (1 + kappa**2)*sin(kappa * x[2]),
                        (1 + kappa**2)*sin(kappa * x[0]))
        f_exact = cf_exact()

    #   3. Refine the mesh to increase the resolution. In this example we do
    #      'ref_levels' of uniform refinement. We choose 'ref_levels' to be the
    #      largest number that gives a final mesh with no more than 50,000
    #      elements.

    ref_levels = int(np.floor(np.log(50000./mesh.GetNE())/np.log(2.)/dim))
    for x in range(ref_levels):
        mesh.UniformRefinement()

    #  4. Define a finite element space on the mesh. Here we use the Nedelec
    #     finite elements of the specified order.

    fec = mfem.ND_FECollection(order, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec)

    print("Number of finite element unknowns: " + str(fespace.GetTrueVSize()))

    # 5. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking all
    #    the boundary attributes from the mesh as essential (Dirichlet) and
    #    converting them to a list of true dofs.

    ess_tdof_list = intArray()
    if mesh.bdr_attributes.Size():
        ess_bdr = intArray(mesh.bdr_attributes.Max())
        ess_bdr = intArray([1]*mesh.bdr_attributes.Max())
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

    # 6. Set up the linear form b(.) which corresponds to the right-hand side
    #    of the FEM linear system, which in this case is (f,phi_i) where f is
    #    given by the function f_exact and phi_i are the basis functions in the
    #    finite element fespace.

    b = mfem.LinearForm(fespace)
    dd = mfem.VectorFEDomainLFIntegrator(f_exact)
    b.AddDomainIntegrator(dd)
    b.Assemble()

    # 7. Define the solution vector x as a finite element grid function
    #    corresponding to fespace. Initialize x by projecting the exact
    #    solution. Note that only values from the boundary edges will be used
    #    when eliminating the non-homogeneous boundary condition to modify the
    #    r.h.s. vector b.

    #from mfem.examples.ex3 import E_exact_cb
    x = mfem.GridFunction(fespace)
    x.ProjectCoefficient(E_exact)

    # 8. Set up the bilinear form corresponding to the EM diffusion operator
    #       curl muinv curl + sigma I, by adding the curl-curl and the mass domain
    #       integrators.

    muinv = mfem.ConstantCoefficient(1.0)
    sigma = mfem.ConstantCoefficient(1.0)
    a = mfem.BilinearForm(fespace)
    a.AddDomainIntegrator(mfem.CurlCurlIntegrator(muinv))
    a.AddDomainIntegrator(mfem.VectorFEMassIntegrator(sigma))

    # 9. Assemble the bilinear form and the corresponding linear system,
    #       applying any necessary transformations such as: eliminating boundary
    #       conditions, applying conforming constraints for non-conforming AMR,
    #       static condensation, etc.

    if (static_cond):
        a.EnableStaticCondensation()
    a.Assemble()

    A = mfem.SparseMatrix()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
    # Here, original version calls hegith, which is not
    # defined in the header...!?
    print("Size of linear system: " + str(A.Size()))
    import sys
    sys.stdout.flush()
    # 10. Solve
    M = mfem.GSSmoother(A)
    mfem.PCG(A, M, B, X, 1, 500, 1e-12, 0.0)
    sys.stdout.flush()

    # 11. Recover the solution as a finite element grid function.
    a.RecoverFEMSolution(X, b, x)

    # 12. Compute and print the L^2 norm of the error.
    import sys
    sys.stdout.write("|| E_h - E ||_{L^2} = " +
                     str(x.ComputeL2Error(E_exact))+"\n")

    mesh.Print('refined.mesh', 8)
    x.Save('sol.gf', 8)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex3 (Definite Maxwell Problem)')
    parser.add_argument('-m', '--mesh',
                        default="beam-tet.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument("-f", "--frequency",
                        action='store',
                        type=float,
                        default=1.0,
                        help="Set the frequency for the exact")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        help='Enable GLVis visualization')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument("-pa", "--partial-assembly",
                        action='store_true',
                        help="Enable Partial Assembly.")
    parser.add_argument("-d", "--device",
                        default="cpu", type=str,
                        help="Device configuration string, see Device::Configure().")
    try:
        from numba import jit
        HAS_NUMBA = True
    except ImportError:
        HAS_NUMBA = False
    parser.add_argument("-n", "--numba",
                        default=int(HAS_NUMBA),
                        type=int,
                        help="Use Number compiled coefficient")

    args = parser.parse_args()
    args.numba = bool(args.numba)
    parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization
    device = args.device
    pa = args.partial_assembly
    freq = args.frequency
    numba = args.numba

    run(freq=freq,
        order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        device=device,
        pa=pa,
        numba=numba)



from  mfem._ser.cpointers import *
from  mfem._ser.globals import *
from  mfem._ser.mem_manager import *
from  mfem._ser.device import *
from  mfem._ser.hash import *
from  mfem._ser.array import *
from  mfem._ser.mesh import *
from  mfem._ser.ncmesh import *
from  mfem._ser.handle import *
from  mfem._ser.point import *
from  mfem._ser.segment import *
from  mfem._ser.triangle import *
from  mfem._ser.quadrilateral import *
from  mfem._ser.wedge import *
from  mfem._ser.tetrahedron import *
from  mfem._ser.hexahedron import *
from  mfem._ser.common_functions import *
from  mfem._ser.operators import *
from  mfem._ser.blockoperator import *
from  mfem._ser.blockvector import *
from  mfem._ser.blockmatrix import *
from  mfem._ser.coefficient import *
from  mfem._ser.lininteg import *
from  mfem._ser.socketstream import *
from  mfem._ser.fe_coll import *
from  mfem._ser.vector import *
from  mfem._ser.complex_operator import *
from  mfem._ser.complex_fem import *
from  mfem._ser.fespace import *
from  mfem._ser.linearform import *
from  mfem._ser.bilininteg import *
from  mfem._ser.gridfunc import *
from  mfem._ser.intrules import *
from  mfem._ser.fe import *
from  mfem._ser.ode import *
from  mfem._ser.bilinearform import *
from  mfem._ser.estimators import *
from  mfem._ser.mesh_operators import *
from  mfem._ser.sparsemat import *
from  mfem._ser.densemat import *
from  mfem._ser.solvers import *
from  mfem._ser.sparsesmoothers import *
from  mfem._ser.eltrans import *
from  mfem._ser.geom import *
from  mfem._ser.vertex import *
from  mfem._ser.table import *
from  mfem._ser.element import *
from  mfem._ser.nonlininteg import *
from  mfem._ser.nonlinearform import *
from  mfem._ser.stable3d import *
from  mfem._ser.vtk import *
from  mfem._ser.datacollection import *
from  mfem._ser.io_stream import wFILE, STDOUT
from  mfem._ser.fespacehierarchy import *
from  mfem._ser.multigrid import *
from  mfem._ser.constraints import *
from  mfem._ser.transfer import *
from  mfem._ser.qspace import *
from  mfem._ser.qfunction import *
from  mfem._ser.quadinterpolator import *
from  mfem._ser.quadinterpolator_face import *

from  mfem._ser.fe_base import *
from  mfem._ser.fe_h1 import *
from  mfem._ser.fe_l2 import *
from  mfem._ser.fe_nd import *
from  mfem._ser.fe_rt import *
from  mfem._ser.fe_ser import *
from  mfem._ser.fe_fixed_order import *
from  mfem._ser.fe_pos import *
from  mfem._ser.fe_nurbs import *
from  mfem._ser.doftrans import *
from  mfem._ser.std_vectors import *

from  mfem._ser.submesh import *
from  mfem._ser.transfermap import *

try:
   from  mfem._ser.gslib import *
except:
   pass

import mfem._ser.array as array
import mfem._ser.blockoperator as blockoperator
import mfem._ser.coefficient as coefficient
import mfem._ser.densemat as densemat
import mfem._ser.error as error
import mfem._ser.fe as fe
import mfem._ser.fe_coll as fe_coll
import mfem._ser.fespace as fespace
import mfem._ser.geom as geom
import mfem._ser.gridfunc as gridfunc
import mfem._ser.intrules as intrules
import mfem._ser.mesh as mesh
import mfem._ser.solvers as solvers
import mfem._ser.vector as vector
import mfem._ser.sparsemat as sparsemat

import mfem._ser.tmop_modules as tmop




'''
   MFEM example 7

   See c++ version in the MFEM library for more detail 
'''
from mfem import path
import mfem.ser as mfem
from mfem.ser import intArray
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, exp


elem_type = 1
ref_levels = 2
amr = 0
order = 2
always_snap = False

if elem_type == 1:
    Nvert = 8
    Nelem = 6
else:
    Nvert = 6
    Nelem = 8


mesh = mfem.Mesh(2, Nvert, Nelem, 0, 3)

if elem_type == 0:
    tri_v = [[1.,  0.,  0.], [0.,  1.,  0.], [-1.,  0.,  0.],
             [0., -1.,  0.], [0.,  0.,  1.], [0.,  0., -1.]]
    tri_e = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4],
             [1, 0, 5], [2, 1, 5], [3, 2, 5], [0, 3, 5]]
    for j in range(Nvert):
        mesh.AddVertex(tri_v[j])
    for j in range(Nelem):
        mesh.AddTriangle(tri_e[j], j+1)
    mesh.FinalizeTriMesh(1, 1, True)
else:
    quad_v = [[-1, -1, -1], [+1, -1, -1], [+1, +1, -1], [-1, +1, -1],
              [-1, -1, +1], [+1, -1, +1], [+1, +1, +1], [-1, +1, +1]]

    quad_e = [[3, 2, 1, 0], [0, 1, 5, 4], [1, 2, 6, 5],
              [2, 3, 7, 6], [3, 0, 4, 7], [4, 5, 6, 7]]
    for j in range(Nvert):
        mesh.AddVertex(quad_v[j])
    for j in range(Nelem):
        mesh.AddQuad(quad_e[j], j+1)
    mesh.FinalizeQuadMesh(1, 1, True)

#  Set the space for the high-order mesh nodes.
fec = mfem.H1_FECollection(order, mesh.Dimension())
nodal_fes = mfem.FiniteElementSpace(mesh, fec, mesh.SpaceDimension())
mesh.SetNodalFESpace(nodal_fes)


def SnapNodes(mesh):
    nodes = mesh.GetNodes()
    node = mfem.Vector(mesh.SpaceDimension())

    for i in np.arange(nodes.FESpace().GetNDofs()):
        for d in np.arange(mesh.SpaceDimension()):
            node[d] = nodes[nodes.FESpace().DofToVDof(i, d)]

        node /= node.Norml2()
        for d in range(mesh.SpaceDimension()):
            nodes[nodes.FESpace().DofToVDof(i, d)] = node[d]


#  3. Refine the mesh while snapping nodes to the sphere.
for l in range(ref_levels+1):
    if l > 0:
        mesh.UniformRefinement()
    if (always_snap or l == ref_levels):
        SnapNodes(mesh)
if amr == 1:
    for l in range(5):
        mesh.RefineAtVertex(mfem.Vertex(0, 0, 1))
    SnapNodes(mesh)
elif amr == 2:
    for l in range(4):
        mesh.RandomRefinement(0.5)
    SnapNodes(mesh)


#  4. Define a finite element space on the mesh. Here we use isoparametric
#      finite elements -- the same as the mesh nodes.
fespace = mfem.FiniteElementSpace(mesh, fec)

print("Number of unknowns: " + str(fespace.GetTrueVSize()))

#   5. Set up the linear form b(.) which corresponds to the right-hand side of
#      the FEM linear system, which in this case is (1,phi_i) where phi_i are
#      the basis functions in the finite element fespace.


class analytic_rhs(mfem.PyCoefficient):
    def EvalValue(self, x):
        l2 = np.sum(x**2)
        return 7*x[0]*x[1]/l2


class analytic_solution(mfem.PyCoefficient):
    def EvalValue(self, x):
        l2 = np.sum(x**2)
        return x[0]*x[1]/l2


b = mfem.LinearForm(fespace)
one = mfem.ConstantCoefficient(1.0)
rhs_coef = analytic_rhs()
sol_coef = analytic_solution()
b.AddDomainIntegrator(mfem.DomainLFIntegrator(rhs_coef))
b.Assemble()

#  6. Define the solution vector x as a finite element grid function
#     corresponding to fespace. Initialize x with initial guess of zero.
x = mfem.GridFunction(fespace)
x.Assign(0.0)

#  7. Set up the bilinear form a(.,.) on the finite element space
#     corresponding to the Laplacian operator -Delta, by adding the Diffusion
#     and Mass domain integrators.
a = mfem.BilinearForm(fespace)
a.AddDomainIntegrator(mfem.DiffusionIntegrator(one))
a.AddDomainIntegrator(mfem.MassIntegrator(one))

#  8. Assemble the linear system, apply conforming constraints, etc.
a.Assemble()
A = mfem.OperatorPtr()
B = mfem.Vector()
X = mfem.Vector()
empty_tdof_list = mfem.intArray()
a.FormLinearSystem(empty_tdof_list, x, b, A, X, B)

AA = mfem.OperatorHandle2SparseMatrix(A)

try:
    umf_solver = mfem.UMFPackSolver()
except:
    umf_solver = None

if umf_solver is None:
    M = mfem.GSSmoother(AA)
    mfem.PCG(AA, M, B, X, 1, 200, 1e-12, 0.0)
else:
    umf_solver.Control[mfem.UMFPACK_ORDERING] = mfem.UMFPACK_ORDERING_METIS
    umf_solver.SetOperator(AA)
    umf_solver.Mult(B, X)

# 10. Recover the solution as a finite element grid function.
a.RecoverFEMSolution(X, b, x)

# 12. Compute and print the L^2 norm of the error.
print("L2 norm of error = " + str(x.ComputeL2Error(sol_coef)))

mesh.Print('sphere_refined.mesh', 8)
x.Save('sol.gf', 8)



import mfem.ser as mfem
import io

mesh = mfem.Mesh(3)
fec = mfem.H1_FECollection(1)
fes = mfem.FiniteElementSpace(mesh, fec)
x = mfem.GridFunction(fes)
x.Assign(0.0)

o = io.StringIO()
l1 = mesh.WriteToStream(o)
l2 = x.WriteToStream(o)

v = mfem.Vector([1,2,3])
l3 = v.WriteToStream(o)

print("result length: ", l1, " ", l2)
print('result: ', o.getvalue())



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    
def run_test(): 
    a = mfem.Vector(np.arange(20.0))
    a.Print()
    
    a *= 30.
    a.Print()

    a[-15:].Print()
    a[-15:].Print_HYPRE()    

    a.Print("vector.dat")
    a.Print_HYPRE("vector_hypre.dat")


    x = mfem.VectorPtrArray([a]*3)
    x[2].Print()
    
if __name__=='__main__':
    run_test()
 



'''

   vector_example.py

   demonstrate how to make mfem::Vector and assign values
   from numpy array

'''
import numpy as np
import mfem.ser as mfem
import traceback as tb

## Vector Creation
print("You can create mfem::Vector from numpy array")
vec = np.array([1,2,3,4,5.])
v = mfem.Vector(vec)
v.Print()

print("Please make sure that you are passing float")
vec = np.array([1,2,3,4,5]).astype(float)
v = mfem.Vector(vec)
v.Print()

print("This one does not work")
vec = np.array([1,2,3,4,5])
try:
   v = mfem.Vector(vec)
except:
   tb.print_exc()    

## Setting Vector using mfem::Vector or numpy array
vec = np.array([1,2,3,4,7.])
v.Assign(vec)
v.Print()

v2 = mfem.Vector(np.array([1,2,3,4,10.]))
v.Assign(v2)
v.Print()

print("This does not work (wrong type)")
try:
    vec = np.array([1,2,3,4,3])
    v.Assign(vec)
    v.Print()
except:
   tb.print_exc()    

print("Also, length must be the same")
try:
   vec = np.array([1,2,3,4.])
   v.Assign(vec)
   v.Print()      
except:
   tb.print_exc()





from __future__ import print_function
import os
from os.path import expanduser, join
import sys
import numpy as np
import io
from mfem import path as mfem_path

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
    smyid = '.'+'{:0>6d}'.format(myid)
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    smyid = ''


def check(a, b, msg):
    assert len(a) == len(b), msg
    assert np.sum(np.abs(a - b)) == 0, msg


def check_mesh(m1, m2, msg):
    assert m1.GetNE() == m2.GetNE(), msg
    assert m1.GetNBE() == m2.GetNBE(), msg
    assert m1.GetNV() == m2.GetNV(), msg


def run_test():
    #meshfile = expanduser(join(mfem_path, 'data', 'semi_circle.mesh'))
    dir = os.path.dirname(os.path.abspath(__file__))    
    meshfile = os.path.join(dir, "../data/amr-quad.mesh")
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    fec = mfem.H1_FECollection(1, dim)
    fespace = mfem.FiniteElementSpace(mesh, fec, 1)
    print('Number of finite element unknowns: ' +
          str(fespace.GetTrueVSize()))

    c = mfem.ConstantCoefficient(1.0)
    gf = mfem.GridFunction(fespace)
    gf.ProjectCoefficient(c)

    odata = gf.GetDataArray().copy()

    gf.Save("out_test_gz"+smyid+".gf")
    gf2 = mfem.GridFunction(mesh, "out_test_gz"+smyid+".gf")
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, "text file does not agree with original")

    gf.Save("out_test_gz"+smyid+".gz")
    gf2.Assign(0.0)
    gf2 = mfem.GridFunction(mesh, "out_test_gz"+smyid+".gz")
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, ".gz file does not agree with original")

    gf.Print("out_test_gz"+smyid+".dat")
    gf2.Assign(0.0)
    gf2.Load("out_test_gz"+smyid+".dat", gf.Size())
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, ".dat file does not agree with original")

    gf.Print("out_test_gz.dat"+smyid+".gz")
    gf2.Assign(0.0)
    gf2.Load("out_test_gz.dat"+smyid+".gz", gf.Size())
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, ".dat file does not agree with original (gz)")

    import gzip
    import io
    gf.Print("out_test_gz.dat2"+smyid+".gz")
    with gzip.open("out_test_gz.dat2"+smyid+".gz", 'rt') as f:
        sio = io.StringIO(f.read())
    gf3 = mfem.GridFunction(fespace)
    gf3.Load(sio, gf.Size())
    odata3 = gf3.GetDataArray().copy()
    check(odata, odata3, ".dat file does not agree with original(gz-io)")

    c = mfem.ConstantCoefficient(2.0)
    gf.ProjectCoefficient(c)
    odata = gf.GetDataArray().copy()

    o = io.StringIO()
    gf.Print(o)
    gf2.Load(o, gf.Size())
    odata2 = gf2.GetDataArray().copy()
    check(odata, odata2, "StringIO does not agree with original")

    print("GridFunction .gf, .gz .dat and StringIO agree with original")

    mesh2 = mfem.Mesh()
    mesh.Print("out_test_gz.mesh")
    mesh2.Load("out_test_gz.mesh")
    check_mesh(mesh, mesh2, ".mesh does not agree with original")

    mesh2 = mfem.Mesh()
    mesh.Print("out_test_gz.mesh"+smyid+".gz")
    mesh2.Load("out_test_gz.mesh"+smyid+".gz")

    check_mesh(mesh, mesh2, ".mesh.gz does not agree with original")

    mesh3 = mfem.Mesh()    
    mesh.PrintGZ("out_test_gz3"+smyid+".mesh")
    mesh3.Load("out_test_gz3"+smyid+".mesh")


    check_mesh(mesh, mesh3, ".mesh (w/o .gz exntension) does not agree with original")
    

    o = io.StringIO()        
    mesh2 = mfem.Mesh()    

    mesh.Print(o)
    mesh2.Load(o)
    check_mesh(mesh, mesh2, ".mesh.gz does not agree with original")

    print("Mesh .mesh, .mesh.gz and StringIO agree with original")

    print("PASSED")


if __name__ == '__main__':
    run_test()



'''
   MFEM example 27p
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt, pi, abs, array, floor, log, arcsin

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def generate_serial_mesh(ref, a_):

    mesh = mfem.Mesh(2, 29, 16, 24, 2)

    for i in range(2):
        o = 13 * i
        mesh.AddQuad([o + 0, o + 3, o + 4, o + 1])
        mesh.AddQuad([o + 1, o + 4, o + 5, o + 2])
        mesh.AddQuad([o + 5, o + 8, o + 9, o + 2])
        mesh.AddQuad([o + 8, o + 12, o + 15, o + 9])
        mesh.AddQuad([o + 11, o + 14, o + 15, o + 12])
        mesh.AddQuad([o + 10, o + 13, o + 14, o + 11])
        mesh.AddQuad([o + 6, o + 13, o + 10, o + 7])
        mesh.AddQuad([o + 0, o + 6, o + 7, o + 3])

    mesh.AddBdrSegment([0, 6], 1)
    mesh.AddBdrSegment([6, 13], 1)
    mesh.AddBdrSegment([13, 19], 1)
    mesh.AddBdrSegment([19, 26], 1)

    mesh.AddBdrSegment([28, 22], 2)
    mesh.AddBdrSegment([22, 15], 2)
    mesh.AddBdrSegment([15, 9], 2)
    mesh.AddBdrSegment([9, 2], 2)

    for i in range(2):
        o = 13 * i
        mesh.AddBdrSegment([o+7, o+3], 3 + i)
        mesh.AddBdrSegment([o+10, o+7], 3 + i)
        mesh.AddBdrSegment([o+11, o+10], 3 + i)
        mesh.AddBdrSegment([o+12, o+11], 3 + i)
        mesh.AddBdrSegment([o+8, o+12], 3 + i)
        mesh.AddBdrSegment([o+5, o+8], 3 + i)
        mesh.AddBdrSegment([o+4, o+5], 3 + i)
        mesh.AddBdrSegment([o+3, o+4], 3 + i)

    a = a_ / sqrt(2)
    mesh.AddVertex([-1.0, -0.5])
    mesh.AddVertex([-1.0, 0.0])
    mesh.AddVertex([-1.0, 0.5])

    mesh.AddVertex([-0.5 - a, -a])
    mesh.AddVertex([-0.5 - a, 0.0])
    mesh.AddVertex([-0.5 - a, a])

    mesh.AddVertex([-0.5, -0.5])
    mesh.AddVertex([-0.5, -a])
    mesh.AddVertex([-0.5, a])
    mesh.AddVertex([-0.5, 0.5])

    mesh.AddVertex([-0.5 + a, -a])
    mesh.AddVertex([-0.5 + a, 0.0])
    mesh.AddVertex([-0.5 + a, a])

    mesh.AddVertex([.0, -0.5])
    mesh.AddVertex([.0, 0.0])
    mesh.AddVertex([.0, 0.5])

    mesh.AddVertex([0.5 - a, -a])
    mesh.AddVertex([0.5 - a, 0.0])
    mesh.AddVertex([0.5 - a, a])

    mesh.AddVertex([0.5, -0.5])
    mesh.AddVertex([0.5, -a])
    mesh.AddVertex([0.5, a])
    mesh.AddVertex([0.5, 0.5])

    mesh.AddVertex([0.5 + a, -a])
    mesh.AddVertex([0.5 + a, 0.0])
    mesh.AddVertex([0.5 + a, a])

    mesh.AddVertex([1.0, -0.5])
    mesh.AddVertex([1.0, 0.0])
    mesh.AddVertex([1.0, 0.5])

    mesh.FinalizeTopology()

    mesh.SetCurvature(1, True)

    # Stitch the ends of the stack together
    # In Python, we just fill list like this
    v2v = [i for i in range(mesh.GetNV() - 3)]
    v2v.append(0)
    v2v.append(1)
    v2v.append(2)

    # renumber elements
    for i in range(mesh.GetNE()):
        el = mesh.GetElement(i)
        nv = el.GetNVertices()
        # We need to re-write the vertex data.
        # el.GetVertices returns *int. So we put it into intArray
        # to access its element
        v = mfem.intArray([el.GetVertices(), nv])
        for j in range(nv):
            v[j] = v2v[v[j]]

          #  renumber boundary elements
        for i in range(mesh.GetNBE()):
            el = mesh.GetBdrElement(i)
            nv = el.GetNVertices()
            v = mfem.intArray([el.GetVertices(), nv])
            for j in range(nv):
                v[j] = v2v[v[j]]

    mesh.RemoveUnusedVertices()
    mesh.RemoveInternalBoundaries()

    mesh.SetCurvature(3, True)

    for l in range(ref):
        mesh.UniformRefinement()

    sdim = mesh.SpaceDimension()

    def quad_trans(u, v, write=False):
        a = a_
        d = 4.0 * a * (sqrt(2) - 2 * a) * (1.0 - 2.0 * v)
        v0 = ((1.0 + sqrt(2)) * (sqrt(2) * a - 2.0 * v) *
              ((4.0 - 3 * sqrt(2)) * a +
               (8.0 * (sqrt(2) - 1.0) * a - 2.0) * v) / d)

        r = (2.0 * ((sqrt(2) - 1.0) * a * a * (1.0 - 4.0 * v) +
                    2.0 * (1.0 + sqrt(2) *
                           (1.0 + 2.0 * (2.0 * a - sqrt(2) - 1.0) * a)) * v * v
                    ) / d)

        t = arcsin(v / r) * u / v
        if write:
            print("u, v, r, v0, t " +
                  "{:g}".format(u) + " " +
                  "{:g}".format(v) + " " +
                  "{:g}".format(r) + " " +
                  "{:g}".format(v0) + " " +
                  "{:g}".format(t))

        x = r * sin(t)
        y = r * cos(t) - v0

        return x, y

    class cTrans(mfem.VectorPyCoefficient):
        def __init__(self):
            mfem.VectorPyCoefficient.__init__(self, sdim)

        def EvalValue(self, u):
            tol = 1e-4
            x = u*0
            if u[1] > 0.5 - tol or u[1] < -0.5 + tol:
                return u
            if (u[0] > 1.0 - tol or u[0] < -1.0 + tol or abs(u[0]) < tol):
                return u
            if u[0] > 0.0:
                if u[1] > abs(u[0] - 0.5):
                    x0, x1 = quad_trans(u[0] - 0.5, u[1])
                    x[0] = x0 + 0.5
                    x[1] = x1
                    return x
                if u[1] < -abs(u[0] - 0.5):
                    x0, x1 = quad_trans(u[0] - 0.5, -u[1])
                    x[0] = x0 + 0.5
                    x[1] = -x1
                    return x
                if u[0] - 0.5 > abs(u[1]):
                    x1, x0 = quad_trans(u[1], u[0] - 0.5)
                    x[0] = x0+0.5
                    x[1] = x1
                    return x
                if u[0] - 0.5 < -abs(u[1]):
                    x1, x0 = quad_trans(u[1], 0.5 - u[0])
                    x[0] = -x0
                    x[0] = x[0] + 0.5
                    x[1] = x1
                    return x
                else:
                    pass
            else:
                if u[1] > abs(u[0] + 0.5):
                    x0, x1 = quad_trans(u[0] + 0.5, u[1])
                    x[0] = x0 - 0.5
                    x[1] = x1
                    return x
                if u[1] < -abs(u[0] + 0.5):
                    x0, x1 = quad_trans(u[0] + 0.5, -u[1])
                    x[0] = x0 - 0.5
                    x[1] = -x1
                    return x
                if u[0] + 0.5 > abs(u[1]):
                    x1, x0 = quad_trans(u[1], u[0] + 0.5)
                    x[0] = x0 - 0.5
                    x[1] = x1
                    return x
                if u[0] + 0.5 < -abs(u[1]):
                    x1, x0 = quad_trans(u[1], -0.5 - u[0])
                    x[0] = -x0
                    x[0] = x[0] - 0.5
                    x[1] = x1
                    return x
            x = u
            return x

    mesh.Transform(cTrans())

    return mesh


def IntegrateBC(x, bdr, alpha, beta, gamma):
    nrm = 0.0
    avg = 0.0
    err = 0.0

    a_is_zero = (alpha == 0.0)
    b_is_zero = (beta == 0.0)

    fes = x.ParFESpace()
    assert fes.GetVDim() == 1, ""

    mesh = fes.GetParMesh()
    shape = mfem.Vector()
    loc_dofs = mfem.Vector()
    w_nor = mfem.Vector()
    dshape = mfem.DenseMatrix()

    dof_ids = mfem.intArray()

    battrs = mesh.GetBdrAttributeArray()

    for i in range(mesh.GetNBE()):
        if bdr[battrs[i]-1] == 0:
            continue

        FTr = mesh.GetBdrFaceTransformations(i)
        if FTr is None:
            continue

        fe = fes.GetFE(FTr.Elem1No)
        assert fe.GetMapType() == mfem.FiniteElement.VALUE, ""

        int_order = 2*fe.GetOrder() + 3

        ir = mfem.IntRules.Get(FTr.GetGeometryType(), int_order)

        dof_ids = fes.GetElementDofs(FTr.Elem1No)

        x.GetSubVector(mfem.intArray(dof_ids), loc_dofs)
        if not a_is_zero:
            sdim = FTr.Face.GetSpaceDim()
            w_nor.SetSize(sdim)
            dshape.SetSize(fe.GetDof(), sdim)
        if not b_is_zero:
            shape.SetSize(fe.GetDof())

        for j in range(ir.GetNPoints()):
            ip = ir.IntPoint(j)
            eip = mfem.IntegrationPoint()
            FTr.Loc1.Transform(ip, eip)
            FTr.Face.SetIntPoint(ip)
            face_weight = FTr.Face.Weight()
            val = 0.0
            if not a_is_zero:
                FTr.Elem1.SetIntPoint(eip)
                fe.CalcPhysDShape(FTr.Elem1, dshape)
                mfem.CalcOrtho(FTr.Face.Jacobian(), w_nor)
                val += alpha * \
                    dshape.InnerProduct(w_nor, loc_dofs) / face_weight
            if not b_is_zero:
                fe.CalcShape(eip, shape)
                val += beta * (shape * loc_dofs)

            # Measure the length of the boundary
            nrm += ip.weight * face_weight

            # Integrate alpha * n.Grad(x) + beta * x
            avg += val * ip.weight * face_weight

            # Integrate |alpha * n.Grad(x) + beta * x - gamma|^2
            val -= gamma
            err += (val*val) * ip.weight * face_weight

        # Normalize by the length of the boundary
    glb_nrm = MPI.COMM_WORLD.allreduce(nrm, op=MPI.SUM)
    glb_avg = MPI.COMM_WORLD.allreduce(avg, op=MPI.SUM)
    glb_err = MPI.COMM_WORLD.allreduce(err, op=MPI.SUM)

    if abs(glb_nrm) > 0.0:
        glb_err /= glb_nrm
        glb_avg /= glb_nrm

    # Compute l2 norm of the error in the boundary condition (negative
    # quadrature weights may produce negative 'err')
    glb_err = sqrt(abs(glb_err))

    # Return the average value of alpha * n.Grad(x) + beta * x
    return glb_err, glb_avg


def run(order=1,
        h1=True,
        sigma=-1.0,
        kappa=-1.0,
        ser_ref_levels=2,
        par_ref_levels=1,
        mat_val=1.0,
        dbc_val=0.0,
        nbc_val=1.0,
        rbc_a_val=1.0,
        rbc_b_val=1.0,
        a_=0.2,
        visualization=True):

    device = mfem.Device('cpu')
    if myid == 0:
        device.Print()

    if kappa < 0 and not h1:
        kappa = (order+1.0)**2
    if a_ < 0.01:
        print("Hole radius too small, resetting to 0.01.")
        a_ = 0.01
    if a_ > 0.49:
        print("Hole radius too large, resetting to 0.49.")
        a_ = 0.49

    # 2. Construct the (serial) mesh and refine it if requested.
    mesh = generate_serial_mesh(ser_ref_levels, a_)
    dim = mesh.Dimension()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    for i in range(par_ref_levels):
        pmesh.UniformRefinement()

    # 3. Define a finite element space on the serial mesh. Here we use either
    #    continuous Lagrange finite elements or discontinuous Galerkin finite
    #   elements of the specified order.
    fec = (mfem.H1_FECollection(order, dim) if h1 else
           mfem.DG_FECollection(order, dim))
    fespace = mfem.ParFiniteElementSpace(pmesh, fec)
    size = fespace.GlobalTrueVSize()
    if myid == 0:
        print("Number of finite element unknowns: " + str(size))

    # 4. Create "marker arrays" to define the portions of boundary associated
    #    with each type of boundary condition. These arrays have an entry
    #    corresponding to each boundary attribute.  Placing a '1' in entry i
    #    marks attribute i+1 as being active, '0' is inactive.
    nbc_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
    rbc_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
    dbc_bdr = mfem.intArray(pmesh.bdr_attributes.Max())

    nbc_bdr.Assign(0)
    nbc_bdr[0] = 1
    rbc_bdr.Assign(0)
    rbc_bdr[1] = 1
    dbc_bdr.Assign(0)
    dbc_bdr[2] = 1

    ess_tdof_list = mfem.intArray()
    if h1 and pmesh.bdr_attributes.Size():
        # For a continuous basis the linear system must be modified to enforce an
        # essential (Dirichlet) boundary condition. In the DG case this is not
        # necessary as the boundary condition will only be enforced weakly.
        fespace.GetEssentialTrueDofs(dbc_bdr, ess_tdof_list)

    # 5. Setup the various coefficients needed for the Laplace operator and the
    #    various boundary conditions. In general these coefficients could be
    #    functions of position but here we use only constants.
    matCoef = mfem.ConstantCoefficient(mat_val)
    dbcCoef = mfem.ConstantCoefficient(dbc_val)
    nbcCoef = mfem.ConstantCoefficient(nbc_val)
    rbcACoef = mfem.ConstantCoefficient(rbc_a_val)
    rbcBCoef = mfem.ConstantCoefficient(rbc_b_val)

    # Since the n.Grad(u) terms arise by integrating -Div(m Grad(u)) by parts we
    # must introduce the coefficient 'm' into the boundary conditions.
    # Therefore, in the case of the Neumann BC, we actually enforce m n.Grad(u)
    # = m g rather than simply n.Grad(u) = g.
    m_nbcCoef = mfem.ProductCoefficient(matCoef, nbcCoef)
    m_rbcACoef = mfem.ProductCoefficient(matCoef, rbcACoef)
    m_rbcBCoef = mfem.ProductCoefficient(matCoef, rbcBCoef)

    # 6. Define the solution vector u as a finite element grid function
    #    corresponding to fespace. Initialize u with initial guess of zero.
    u = mfem.ParGridFunction(fespace)
    u.Assign(0.0)

    # 7. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the Laplacian operator -Delta, by adding the Diffusion
    #    domain integrator.
    a = mfem.ParBilinearForm(fespace)
    a.AddDomainIntegrator(mfem.DiffusionIntegrator(matCoef))
    if h1:
        # Add a Mass integrator on the Robin boundary
        a.AddBoundaryIntegrator(mfem.MassIntegrator(m_rbcACoef), rbc_bdr)
    else:
        # Add the interfacial portion of the Laplace operator
        a.AddInteriorFaceIntegrator(mfem.DGDiffusionIntegrator(matCoef,
                                                               sigma, kappa))

        # Counteract the n.Grad(u) term on the Dirichlet portion of the boundary
        a.AddBdrFaceIntegrator(mfem.DGDiffusionIntegrator(matCoef, sigma, kappa),
                               dbc_bdr)

        # Augment the n.Grad(u) term with a*u on the Robin portion of boundary
        a.AddBdrFaceIntegrator(
            mfem.BoundaryMassIntegrator(m_rbcACoef), rbc_bdr)
    a.Assemble()

    # 8. Assemble the linear form for the right hand side vector.
    b = mfem.ParLinearForm(fespace)

    if h1:
        # Set the Dirichlet values in the solution vector
        u.ProjectBdrCoefficient(dbcCoef, dbc_bdr)

        # Add the desired value for n.Grad(u) on the Neumann boundary
        b.AddBoundaryIntegrator(mfem.BoundaryLFIntegrator(m_nbcCoef), nbc_bdr)

        # Add the desired value for n.Grad(u) + a*u on the Robin boundary
        b.AddBoundaryIntegrator(mfem.BoundaryLFIntegrator(m_rbcBCoef), rbc_bdr)
    else:
        # Add the desired value for the Dirichlet boundary
        b.AddBdrFaceIntegrator(mfem.DGDirichletLFIntegrator(dbcCoef, matCoef,
                                                            sigma, kappa),
                               dbc_bdr)

        # Add the desired value for n.Grad(u) on the Neumann boundary
        b.AddBdrFaceIntegrator(mfem.BoundaryLFIntegrator(m_nbcCoef), nbc_bdr)

        # Add the desired value for n.Grad(u) + a*u on the Robin boundary
        b.AddBdrFaceIntegrator(mfem.BoundaryLFIntegrator(m_rbcBCoef), rbc_bdr)

    b.Assemble()

    # 9. Construct the linear system.
    A = mfem.OperatorPtr()
    B = mfem.Vector()
    X = mfem.Vector()
    a.FormLinearSystem(ess_tdof_list, u, b, A, X, B)

    # 10. Define a simple symmetric Gauss-Seidel preconditioner and use it to
    #     solve the system AX=B with PCG in the symmetric case, and GMRES in the
    #     non-symmetric one.
    amg = mfem.HypreBoomerAMG()
    if sigma == -1.0:
        pcg = mfem.HyprePCG(MPI.COMM_WORLD)
        pcg.SetTol(1e-12)
        pcg.SetMaxIter(200)
        pcg.SetPrintLevel(2)
        pcg.SetPreconditioner(amg)
        pcg.SetOperator(A.Ptr())
        pcg.Mult(B, X)
    else:
        gmres.GMRESSolver(MPI.COMM_WORLD)
        gmres.SetAbsTol(0.0)
        gmres.SetRelTol(1e-12)
        gmres.SetMaxIter(200)
        gmres.SetKDim(10)
        gmres.SetPrintLevel(1)
        gmres.SetPreconditioner(amg)
        gmres.SetOperator(A.Ptr())
        gmres.Mult(B, X)

    # 12. Recover the grid function corresponding to U. This is the local finite
    #     element solution.
    a.RecoverFEMSolution(X, b, u)

    # 13. Build a mass matrix to help solve for n.Grad(u) where 'n' is a surface
    #     normal.
    m = mfem.ParBilinearForm(fespace)
    m.AddDomainIntegrator(mfem.MassIntegrator())
    m.Assemble()

    ess_tdof_list.SetSize(0)
    M = mfem.OperatorPtr()
    m.FormSystemMatrix(ess_tdof_list, M)

    # 14. Compute the various boundary integrals.
    if myid == 0:
        print("Verifying boundary conditions" +
              "=============================")

    # Integrate the solution on the Dirichlet boundary and compare to the
    # expected value.
    def print0(*args):
        if myid == 0:
            print(*args)

    err, avg = IntegrateBC(u, dbc_bdr, 0.0, 1.0, dbc_val)

    hom_dbc = (dbc_val == 0.0)
    err /= 1.0 if hom_dbc else abs(dbc_val)
    print0("Average of solution on Gamma_dbc:\t" +
           "{:g}".format(avg) + ", \t" +
           ("absolute" if hom_dbc else "relative") +
           " error " + "{:g}".format(err))

    # Integrate n.Grad(u) on the inhomogeneous Neumann boundary and compare
    # to the expected value.
    err, avg = IntegrateBC(u, nbc_bdr, 1.0, 0.0, nbc_val)

    hom_nbc = (nbc_val == 0.0)
    err /= 1.0 if hom_nbc else abs(nbc_val)
    print0("Average of n.Grad(u) on Gamma_ndbc:\t" +
           "{:g}".format(avg) + ", \t" +
           ("absolute" if hom_nbc else "relative") +
           " error " + "{:g}".format(err))

    # Integrate n.Grad(u) on the homogeneous Neumann boundary and compare to
    # the expected value of zero.
    nbc0_bdr = mfem.intArray(mesh.bdr_attributes.Max())
    nbc0_bdr.Assign(0)
    nbc0_bdr[3] = 1

    err, avg = IntegrateBC(u, nbc0_bdr, 1.0, 0.0, 0.0)
    hom_nbc = True
    print0("Average of n.Grad(u) on Gamma_ndbc0:\t" +
           "{:g}".format(avg) + ", \t" +
           ("absolute" if hom_nbc else "relative") +
           " error " + "{:g}".format(err))

    # Integrate n.Grad(u) + a * u on the Robin boundary and compare to the
    # expected value.
    err, avg = IntegrateBC(u, rbc_bdr, 1.0, rbc_a_val, rbc_b_val)

    hom_rbc = (rbc_b_val == 0.0)
    err /= 1.0 if hom_rbc else abs(nbc_val)
    print0("Average of n.Grad(u)+a*u on Gamma_rdbc:\t" +
           "{:g}".format(avg) + ", \t" +
           ("absolute" if hom_rbc else "relative") +
           " error " + "{:g}".format(err))

    # 15. Save the refined mesh and the solution. This output can be viewed
    #     later using GLVis: "glvis -m refined.mesh -g sol.gf".
    pmesh.Print("mesh"+smyid, 8)
    u.Save("sol"+smyid, 8)

    # 16. Send the solution by socket to a GLVis server.
    if visualization:
        title_str = "H1" if h1 else "DG"
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock << "parallel " << num_procs << " " << myid << "\n"
        sol_sock.precision(8)
        sol_sock << "solution\n" << mesh << u
        sol_sock << "window_title '" << title_str << " Solution'"
        sol_sock << " keys 'mmc'"
        sol_sock.flush()


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex27 (Laplace boundary conditionss)')

    parser.add_argument("-h1", "--continuous",
                        action='store_true', default=True,
                        help='Select continuous "H1" element')
    parser.add_argument("-dg", "--discontinuous",
                        action='store_true', default=False,
                        help='Select continuous "DG" element')
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument("-s", "--sigma",
                        action='store', default=-1.0, type=float,
                        help="One of the two DG penalty parameters, typically +1/-1." +
                        " See the documentation of class DGDiffusionIntegrator.")
    parser.add_argument("-k", "--kappa",
                        action='store', default=-1.0, type=float,
                        help="One of the two DG penalty parameters, should be positive." +
                        " Negative values are replaced with (order+1)^2.")
    parser.add_argument("-rs", "--refine-serial",
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly in serial.")
    parser.add_argument("-rp", "--refine-parallel",
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly in parallel.")
    parser.add_argument("-mat", "--material-value",
                        action='store', default=1.0, type=float,
                        help="Constant value for material coefficient " +
                        "in the Laplace operator.")
    parser.add_argument("-dbc", "--dirichlet-value",
                        action='store', default=0.0, type=float,
                        help="Constant value for Dirichlet Boundary Condition.")
    parser.add_argument("-nbc", "--neumann-value",
                        action='store', default=1.0, type=float,
                        help="Constant value for Neumann Boundary Condition.")
    parser.add_argument("-rbc-a", "--robin-a-value",
                        action='store', default=1.0, type=float,
                        help="Constant 'a' value for Robin Boundary Condition. " +
                        "du/dn + a * u = b.")
    parser.add_argument("-rbc-b", "--robin-b-value",
                        action='store', default=1.0, type=float,
                        help="Constant 'b' value for Robin Boundary Condition: " +
                        "du/dn + a * u = b.")
    parser.add_argument("-a", "--radius",
                        action='store', default=0.2, type=float,
                        help="Radius of holes in the mesh.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true', default=True,
                        help='Enable GLVis visualization')
    parser.add_argument('-no-vis', '--no_visualization',
                        action='store_true', default=False,
                        help='Disable GLVis visualization')

    args = parser.parse_args()

    h1 = args.continuous
    if args.discontinuous:
        h1 = False
        args.continuous = False
    vis = True
    if args.no_visualization:
        vis = False
        args.visualization = False

    if myid == 0:
        parser.print_options(args)
    run(order=args.order,
        h1=h1,
        sigma=args.sigma,
        kappa=args.kappa,
        ser_ref_levels=args.refine_serial,
        par_ref_levels=args.refine_parallel,
        mat_val=args.material_value,
        dbc_val=args.dirichlet_value,
        nbc_val=args.neumann_value,
        rbc_a_val=args.robin_a_value,
        rbc_b_val=args.robin_b_value,
        a_=args.radius,
        visualization=vis)



'''
 This example shows
   constructing HYPREParCSR in a distributed manner
     1) build scipy.sparse on each node
           each node has informaiton of small number of rows
     2) assemble it to HYPREParCSR
           scipy matries are concatenated vertical

   convert back ParCSR to scipy.sparse
     1) on each node, it produce scipy.sparse which contain
        only its own rows.


   example_2 tests
      concatenate 2 of 2x5 matrices and one 1x5 to form 5x5
      peform ParMMULTI, RAP, and ParAdd 
      works only for mpirun -np 3 (3 cpu)

'''
import mfem.par as par
from mfem.common.parcsr_extra import *

from scipy.sparse import csr_matrix, coo_matrix

from mpi4py import MPI
comm     = MPI.COMM_WORLD     
num_proc = MPI.COMM_WORLD.size
myid     = MPI.COMM_WORLD.rank

def print_hypre(M, txt):
    for i in range(num_proc):
       MPI.COMM_WORLD.Barrier()                              
       if myid == i:
          if myid == 0:
              print(txt)
              print('MyID: ', myid)
          else:
              print('MyID: ', myid)
          print(ToScipyCoo(M))

shape = (2, 5) if myid < 2 else (1, 5)
# make sample matrix
row  = np.array([0, 0, 0, 0])
col  = np.array([0, 3, 1, 4])
data = np.array([4, 5, 7, 9])
m = coo_matrix((data, (row, col)), shape=shape)
m = m.tocsr()
m = m*(myid+1)

M = ToHypreParCSR(m)
print_hypre(M, 'matrix M')

# make sample matrix
if True:
   row  = np.array([0, 0, 0, 0])
   col  = np.array([4, 3, 1, 2])
   data = np.array([4, 10,7, 2 ])
   m = coo_matrix((data, (row, col)), shape=shape).tocsr()
   M2 = ToHypreParCSR(m)
else:
   m2 = ToScipyCoo(M)
   M2 = ToHypreParCSR(m2.tocsr())
#print ToScipyCoo(M2)

# adding matrix
M3 = ParAdd(M, M2)


print_hypre(M3, 'summed matrix')
print_hypre(mfem.ParMult(M, M2.Transpose()), 'parmult (A*B)') #ok
print_hypre(mfem.RAP(M, M2.Transpose()), 'rap (B A Bt)')      #ok
print_hypre(mfem.ParMult(M, M2), 'parmult (A*Bt)')            #ok
print_hypre(mfem.RAP(M, M2), 'rap (Bt A B)')                  #ok






'''
   MFEM example 21
      See c++ version in the MFEM library for more detail 
'''
import os
import mfem.par as mfem
from mfem.par import intArray
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sin, cos, exp, sqrt

from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(order=1,
        static_cond=False,
        meshfile="",
        visualization=1,
        serial_ref_levels=0):

    device = mfem.Device('cpu')
    if myid == 0:
        device.Print()

    # 2. Read the mesh from the given mesh file. We can handle triangular,
    #    quadrilateral, tetrahedral, and hexahedral meshes with the same code.
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    assert mesh.SpaceDimension() == dim, "invalid mesh"

    if mesh.attributes.Max() < 2 or mesh.bdr_attributes.Max() < 2:
        print("\n".join(["Input mesh should have at least two materials and "
                         "two boundary attributes! (See schematic in ex2.cpp)"]))

    # 3. Since a NURBS mesh can currently only be refined uniformly, we need to
    #    convert it to a piecewise-polynomial curved mesh. First we refine the
    #    NURBS mesh a bit more and then project the curvature to quadratic Nodes.
    if mesh.NURBSext and serial_ref_levels == 0:
        serial_ref_levels = 2

    for i in range(serial_ref_levels):
        mesh.UniformRefinement()

    if mesh.NURBSext:
        mesh.SetCurvature(2)
    mesh.EnsureNCMesh()

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)

    # 4. Define a finite element space on the mesh. The polynomial order is
    #    one (linear) by default, but this can be changed on the command line.
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec, dim)

    # 5. As in Example 2, we set up the linear form b(.) which corresponds to
    #    the right-hand side of the FEM linear system. In this case, b_i equals
    #    the boundary integral of f*phi_i where f represents a "pull down"
    #    force on the Neumann part of the boundary and phi_i are the basis
    #    functions in the finite element fespace. The force is defined by the
    #    VectorArrayCoefficient object f, which is a vector of Coefficient
    #    objects. The fact that f is non-zero on boundary attribute 2 is
    #    indicated by the use of piece-wise constants coefficient for its last
    #    component. We don't assemble the discrete problem yet, this will be
    #    done in the main loop.
    f = mfem.VectorArrayCoefficient(dim)
    for i in range(dim-1):
        f.Set(i, mfem.ConstantCoefficient(0.0))

    pull_force = mfem.Vector(pmesh.bdr_attributes.Max())
    pull_force.Assign(0.0)
    pull_force[1] = -1.0e-2
    f.Set(dim-1, mfem.PWConstCoefficient(pull_force))

    b = mfem.ParLinearForm(fespace)
    b.AddDomainIntegrator(mfem.VectorBoundaryLFIntegrator(f))

    # 6. Set up the bilinear form a(.,.) on the finite element space
    #    corresponding to the linear elasticity integrator with piece-wise
    #    constants coefficient lambda and mu.
    llambda = mfem.Vector(pmesh.attributes.Max())
    llambda.Assign(1.0)
    llambda[0] = llambda[1]*50
    lambda_func = mfem.PWConstCoefficient(llambda)
    mu = mfem.Vector(pmesh.attributes.Max())
    mu.Assign(1.0)
    mu[0] = mu[1]*50
    mu_func = mfem.PWConstCoefficient(mu)

    a = mfem.ParBilinearForm(fespace)
    integ = mfem.ElasticityIntegrator(lambda_func, mu_func)
    a.AddDomainIntegrator(integ)
    if static_cond:
        a.EnableStaticCondensation()

    # 7. The solution vector x and the associated finite element grid function
    #    will be maintained over the AMR iterations. We initialize it to zero.
    zero_vec = mfem.Vector(dim)
    zero_vec.Assign(0.0)
    zero_vec_coeff = mfem.VectorConstantCoefficient(zero_vec)
    x = mfem.ParGridFunction(fespace)
    x.Assign(0.0)

    # 8. Determine the list of true (i.e. conforming) essential boundary dofs.
    #    In this example, the boundary conditions are defined by marking only
    #    boundary attribute 1 from the mesh as essential and converting it to a
    #    list of true dofs.  The conversion to true dofs will be done in the
    #    main loop.
    ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
    ess_bdr.Assign(0)
    ess_bdr[0] = 1

    # 9. Connect to GLVis.
    if visualization:
        sol_sock = mfem.socketstream("localhost", 19916)
        sol_sock.precision(8)

    # 10. Set up an error estimator. Here we use the Zienkiewicz-Zhu estimator
    #     that uses the ComputeElementFlux method of the ElasticityIntegrator to
    #     recover a smoothed flux (stress) that is subtracted from the element
    #     flux to get an error indicator. We need to supply the space for the
    #     smoothed flux: an (H1)^tdim (i.e., vector-valued) space is used here.
    #     Here, tdim represents the number of components for a symmetric (dim x
    #     dim) tensor.
    tdim = dim*(dim+1)//2
    flux_fec = mfem.L2_FECollection(order, dim)
    flux_fespace = mfem.ParFiniteElementSpace(pmesh, flux_fec, tdim)
    smooth_flux_fespace = mfem.ParFiniteElementSpace(pmesh, fec, tdim)
    estimator = mfem.L2ZienkiewiczZhuEstimator(
        integ, x, flux_fespace, smooth_flux_fespace)

    # 11. A refiner selects and refines elements based on a refinement strategy.
    #     The strategy here is to refine elements with errors larger than a
    #     fraction of the maximum element error. Other strategies are possible.
    #     The refiner will call the given error estimator.
    refiner = mfem.ThresholdRefiner(estimator)
    refiner.SetTotalErrorFraction(0.7)

    # 12. The main AMR loop. In each iteration we solve the problem on the
    #     current mesh, visualize the solution, and refine the mesh.
    max_dofs = 50000
    max_amr_itr = 20
    for it in range(max_amr_itr+1):
        global_dofs = fespace.GlobalTrueVSize()
        if myid == 0:
            print("\n".join(["",
                             "AMR iteration " + str(it),
                             "Number of unknowns: " + str(global_dofs)]))

        # 13. Assemble the stiffness matrix and the right-hand side.
        a.Assemble()
        b.Assemble()

        # 14. Set Dirichlet boundary values in the GridFunction x.
        #     Determine the list of Dirichlet true DOFs in the linear system.
        ess_tdof_list = mfem.intArray()
        x.ProjectBdrCoefficient(zero_vec_coeff, ess_bdr)
        fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)

        # 15. Create the linear system: eliminate boundary conditions, constrain
        #     hanging nodes and possibly apply other transformations. The system
        #     will be solved for true (unconstrained) DOFs only.
        A = mfem.HypreParMatrix()
        X = mfem.Vector()
        B = mfem.Vector()
        copy_interior = 1
        a.FormLinearSystem(ess_tdof_list, x, b, A, X, B, copy_interior)

        # 16. Define and apply a parallel PCG solver for AX=B with the BoomerAMG
        #     preconditioner from hypre.
        amg = mfem.HypreBoomerAMG()
        amg.SetPrintLevel(0)
        # amg.SetSystemsOptions(dim); // optional
        pcg = mfem.CGSolver(A.GetComm())
        pcg.SetPreconditioner(amg)
        pcg.SetOperator(A)
        pcg.SetRelTol(1e-6)
        pcg.SetMaxIter(500)
        pcg.SetPrintLevel(3)  # print the first and the last iterations only
        pcg.Mult(B, X)

        # 17. After solving the linear system, reconstruct the solution as a
        #     finite element GridFunction. Constrained nodes are interpolated
        #     from true DOFs (it may therefore happen that x.Size() >= X.Size()).
        a.RecoverFEMSolution(X, b, x)

        # 18. Send solution by socket to the GLVis server.
        if visualization and sol_sock.good():
            nodes = mfem.GridFunction(fespace)
            pmesh.GetNodes(nodes)
            nodes += x
            own_nodes = 0
            nodes, own_nodes = pmesh.SwapNodes(nodes, own_nodes)

            x.Neg()  # visualize the backward displacement
            sol_sock << "parallel " << num_procs << ' ' << myid << '\n'
            sol_sock << "solution\n" << mesh << x
            sol_sock.flush()
            x.Neg()

            nodes, own_nodes = pmesh.SwapNodes(nodes, own_nodes)
            if it == 0:
                kk = "Rj1" if dim == 2 else ""
                sol_sock << "keys '" << kk << "m'" << "\n"
                sol_sock.endline()

            sol_sock << "window_title 'AMR iteration: " << it << "'\n" << "pause"

            if myid == 0:
                print("Visualization paused. "
                      "Press <space> in the GLVis window to continue.")

        if (global_dofs > max_dofs):
            if myid == 0:
                print("Reached the maximum number of dofs. Stop.")
            break

        # 19. Call the refiner to modify the mesh. The refiner calls the error
        #     estimator to obtain element errors, then it selects elements to be
        #     refined and finally it modifies the mesh. The Stop() method can be
        #     used to determine if a stopping criterion was met.
        refiner.Apply(pmesh)
        if (refiner.Stop()):
            if myid == 0:
                print("Stopping criterion satisfied. Stop.")
            break

        # 20. Update the space to reflect the new state of the mesh. Also,
        #     interpolate the solution x so that it lies in the new space but
        #     represents the same function. This saves solver iterations later
        #     since we'll have a good initial guess of x in the next step.
        #     Internally, FiniteElementSpace::Update() calculates an
        #     interpolation matrix which is then used by GridFunction::Update().
        fespace.Update()
        x.Update()

        # 21. Load balance the mesh, and update the space and solution. Currently
        #     available only for nonconforming meshes.
        if pmesh.Nonconforming():
            pmesh.Rebalance()
            fespace.Update()
            x.Update()
        a.Update()
        b.Update()

    pmesh.Print("ex21p_reference_mesh"+smyid, 16)
    nodes = mfem.GridFunction(fespace)
    pmesh.GetNodes(nodes)
    nodes += x
    own_nodes = 0
    nodes2, own_nodes = pmesh.SwapNodes(nodes, own_nodes)
    pmesh.Print("ex21p_deformed_mesh"+smyid, 16)
    nodes, own_nodes = pmesh.SwapNodes(nodes2, own_nodes)

    x.Save("ex21p_displacement"+smyid)


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(
        description='Ex21 (Adaptive mesh refinement for linear elasticity)')
    parser.add_argument('-m', '--mesh',
                        default='beam-tri.mesh',
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument("-rs", "--refine_serial",
                        action='store', type=int, default=0,
                        help="Number of serial refinement before parallel partitioning")
    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree)")
    parser.add_argument('-sc', '--static-condensation',
                        action='store_true',
                        help="Enable static condensation.")
    parser.add_argument('-vis', '--visualization',
                        action='store_true',
                        default=True,
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    order = args.order
    static_cond = args.static_condensation

    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = args.visualization

    run(order=order,
        static_cond=static_cond,
        meshfile=meshfile,
        visualization=visualization,
        serial_ref_levels=args.refine_serial)



file = "../_ser/lininteg.py"

out = ["namespace mfem {"]
fid = open(file, 'r')
for line in fid.readlines():
    if line.startswith("class"):
        cname = (line.split(' ')[1]).split('(')[0]
         
    if line.startswith("    def __init__"):
        pp = ""
        if line.find("*args") != -1:
            pp = "    self._coeff = args"
        elif line.find(", QG") != -1:
            pp = "    self._coeff = QG"
        elif line.find(", QF)") != -1:
            pp = "    self._coeff = QF"
        elif line.find(", F)") != -1:
            pp = "    self._coeff = F"
        elif line.find(", f, s=1.0, ir=None)") != -1:
            pp = "    self._coeff = (f, ir)"
        elif line.find(", uD_, lambda_, mu_, alpha_, kappa_)") != -1:
            pp = "    self._coeff = uD_"
        elif line.find("(self)") != -1:
            pass
        else:
            print(cname)
            print(line)
            assert False, "No recipt for this pattern "
        if pp != "":
            out.append("%pythonappend " + cname + "::" + cname + " %{")
            out.append(pp)
            out.append("%}")
fid.close()
out.append("}")

fid = open("lininteg_ext.i", "w")
fid.write("\n".join(out))
fid.close()




'''
   MFEM example 32

   See c++ version in the MFEM library for more detail

   Sample runs:  mpirun -np 4 python ex32p.py -m ../data/hexagon.mesh -o 2
                 mpirun -np 4 python ex32p.py -m ../data/star.mesh
                 mpirun -np 4 python ex32p.py -m ../data/square-disc.mesh -o 2 -n 4 -rs 1
                 mpirun -np 4 python ex32p.py -m ../data/square-disc-nurbs.mesh -rs 3 -o 3
                 mpirun -np 4 python ex32p.py -m ../data/amr-quad.mesh -o 2 -rs 1
                 mpirun -np 4 python ex32p.py -m ../data/amr-hex.mesh -rs 1
                 mpirun -np 4 python ex32p.py -m ../data/fichera.mesh -rs 1

'''
from mpi4py import MPI
import mfem.par as mfem
from mfem.par import intArray, doubleArray
import os
import sys
from os.path import expanduser, join
import numpy as np
from numpy import sin, cos, array, pi, sqrt

sqrt1_2 = 1/sqrt(2)
sqrt2 = sqrt(2)

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank
smyid = '.'+'{:0>6d}'.format(myid)


def run(nev=5,
        order=1,
        rs=2,
        rp=1,
        meshfile='',
        visualization=False):

    # 3. Read the (serial) mesh from the given mesh file on all processors. We
    #    can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
    #    and volume meshes with the same code.
    mesh = mfem.Mesh(meshfile, 1, 1)
    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()

    bbMin, bbMax = mesh.GetBoundingBox()

    # 4. Refine the mesh to increase the resolution. In this example we do
    #    'ref_levels' of uniform refinement (2 by default, or specified on
    #     the command line with -rs)
    for lev in range(rs):
        mesh.UniformRefinement()

    # 5. Define a parallel mesh by a partitioning of the serial mesh. Refine
    #    this mesh further in parallel to increase the resolution (1 time by
    #    default, or specified on the command line with -rp). Once the parallel

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    del mesh
    for lev in range(rp):
        pmesh.UniformRefinement()

    # 6. Define a parallel finite element space on the parallel mesh. Here we
    #    use the Nedelec finite elements of the specified order.
    if dim == 1:
        fec_nd = mfem.ND_R1D_FECollection(order, dim)
        fec_rt = mfem.RT_R1D_FECollection(order-1, dim)
    elif dim == 2:
        fec_nd = mfem.ND_R2D_FECollection(order, dim)
        fec_rt = mfem.RT_R2D_FECollection(order-1, dim)
    else:
        fec_nd = mfem.ND_FECollection(order, dim)
        fec_rt = mfem.RT_FECollection(order-1, dim)

    fespace_nd = mfem.ParFiniteElementSpace(pmesh, fec_nd)
    fespace_rt = mfem.ParFiniteElementSpace(pmesh, fec_rt)

    size_nd = fespace_nd.GlobalTrueVSize()
    size_rt = fespace_rt.GlobalTrueVSize()

    if myid == 0:
        print("Number of H(Curl) unknowns: " + str(size_nd))
        print("Number of H(Div) unknowns: " + str(size_rt))

    # 7. Set up the parallel bilinear forms a(.,.) and m(.,.) on the finite
    #    element space. The first corresponds to the curl curl, while the second
    #    is a simple mass matrix needed on the right hand side of the
    #    generalized eigenvalue problem below. The boundary conditions are
    #    implemented by marking all the boundary attributes from the mesh as
    #    essential. The corresponding degrees of freedom are eliminated with
    #    special values on the diagonal to shift the Dirichlet eigenvalues out
    #    of the computational range. After serial and parallel assembly we
    #    extract the corresponding parallel matrices A and M.
    shift = 0.0
    mat = array([[2.0, sqrt1_2, 0, ],
                 [sqrt1_2, 2.0, sqrt1_2],
                 [0.0, sqrt1_2, 2.0, ], ])
    epsilon_mat = mfem.DenseMatrix(mat)
    epsilon = mfem. MatrixConstantCoefficient(epsilon_mat)
    one = mfem.ConstantCoefficient(1.0)

    if pmesh.bdr_attributes.Size():
        ess_bdr = intArray([1]*pmesh.bdr_attributes.Max())

    a = mfem.ParBilinearForm(fespace_nd)
    curlcurl = mfem.CurlCurlIntegrator(one)
    a.AddDomainIntegrator(curlcurl)

    mass = mfem.VectorFEMassIntegrator(epsilon)

    if pmesh.bdr_attributes.Size() == 0 or dim == 1:
        #  Add a mass term if the mesh has no boundary, e.g. periodic mesh or
        #  closed surface.
        a.AddDomainIntegrator(mass)
        shift = 1.0
        if myid == 0:
            print("Computing eigenvalues shifted by " + str(1.0))

    a.Assemble()
    a.EliminateEssentialBCDiag(ess_bdr, 1.0)
    a.Finalize()

    m = mfem.ParBilinearForm(fespace_nd)
    m.AddDomainIntegrator(mass)
    m.Assemble()

    # shift the eigenvalue corresponding to eliminated dofs to a large value
    m.EliminateEssentialBCDiag(ess_bdr, 2.3e-308)
    m.Finalize()

    A = a.ParallelAssemble()
    M = m.ParallelAssemble()

    # 8. Define and configure the AME eigensolver and the AMS preconditioner for
    #    A to be used within the solver. Set the matrices which define the
    #    generalized eigenproblem A x = lambda M x.
    ams = mfem.HypreAMS(A, fespace_nd)
    ams.SetPrintLevel(0)
    ams.SetSingularProblem()

    ame = mfem.HypreAME(MPI.COMM_WORLD)
    ame.SetNumModes(nev)
    ame.SetPreconditioner(ams)
    ame.SetMaxIter(100)
    ame.SetTol(1e-8)
    ame.SetPrintLevel(1)
    ame.SetMassMatrix(M)
    ame.SetOperator(A)

    # 9. Compute the eigenmodes and extract the array of eigenvalues. Define
    #    parallel grid functions to represent each of the eigenmodes returned by
    #    the solver and their derivatives.
    eigenvalues = doubleArray()
    ame.Solve()
    ame.GetEigenvalues(eigenvalues)
    x = mfem.ParGridFunction(fespace_nd)
    dx = mfem.ParGridFunction(fespace_rt)

    curl = mfem.ParDiscreteLinearOperator(fespace_nd, fespace_rt)
    curl.AddDomainInterpolator(mfem.CurlInterpolator())
    curl.Assemble()
    curl.Finalize()

    # 10. Save the refined mesh and the modes in parallel. This output can be
    #     viewed later using GLVis: "glvis -np <np> -m mesh -g mode".
    smyid = '{:0>6d}'.format(myid)
    mesh_name = "mesh."+smyid
    pmesh.Print(mesh_name, 8)

    for i in range(nev):
        x.Assign(ame.GetEigenvector(i))
        curl.Mult(x, dx)

        mode_name = "mode_"+str(i).zfill(2)+"."+smyid
        mode_deriv_name = "mode_deriv_"+str(i).zfill(2)+"."+smyid

        x.Save(mode_name, 8)
        dx.Save(mode_deriv_name, 8)

    # 11 Visualize data using glvis
    if visualization:
        # 11. (a) functions to format Glvis window
        def make_cmd(label, px, py, max_r, i, title="Eigenmode"):
            cmd = (" window_title '" + title + " " + str(i+1) + '/' + str(nev) +
                   " " + label + ", Lambda = " + "{:g}".format(eigenvalues[i] - shift) +
                   "'" + " valuerange -" + str(max_r) + ' ' + str(max_r))
            cmd = (cmd + " keys aa window_geometry " +
                   str(px) + " " + str(py) + " 400 350")
            return cmd
        # 11. (b) fucntion to send data to Glvis

        def send_data_to_glvis(sock, data, cmd):
            sock << "parallel " << num_procs << " " << myid << "\n"
            sock << "solution\n" << pmesh << data
            sock.flush()
            sock << cmd
            sock.endline()
            MPI.COMM_WORLD.Barrier()

        def ask_exit():
            if (myid == 0):
                from builtins import input
                c = input("press (q)uit or (c)ontinue --> ")
            else:
                c = None
            c = MPI.COMM_WORLD.bcast(c, root=0)
            if (c != 'c'):
                sys.exit()

        if dim == 1:
            mode_x_sock = make_socketstrema()
            mode_y_sock = make_socketstrema()
            mode_z_sock = make_socketstrema()
            mode_dy_sock = make_socketstrema()
            mode_dz_sock = make_socketstrema()

            xVec = mdfem.Vector([1, 0, 0.])
            yVec = mdfem.Vector([0, 1, 0.])
            zVec = mdfem.Vector([0, 0, 1.])

            xVecCof = mfem.VectorConstantCoefficient(xVec)
            yVecCof = mfem.VectorConstantCoefficient(yVec)
            zVecCof = mfem.VectorConstantCoefficient(zVec)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.ParFiniteElementSpace(pmesh, fec_h1)
            fes_l2 = mfem.ParFiniteElementSpace(pmesh, fec_l2)

            xComp = mfem.ParGridFunction(fes_l2)
            yComp = mfem.ParGridFunction(fes_h1)
            zComp = mfem.ParGridFunction(fes_h1)

            dyComp = mfem.ParGridFunction(fes_l2)
            dzComp = mfem.ParGridFunction(fes_l2)

            for i in range(nev):
                if (myid == 0):
                    print("Eigenmode " + str(i+1) + '/' + str(nev) +
                          ", Lambda = " + "{:g}".format(eigenvalues[i]))

                x.Assign(ame.GetEigenvector(i))
                curl.Mult(x, dx)

                modeCoeff = mfem.VectorGridFunctionCoefficient(x)
                xCoef = mfem.InnerProductCoefficient(xVecCoef, modeCoef)
                yCoef = mfem.InnerProductCoefficient(yVecCoef, modeCoef)
                zCoef = mfem.InnerProductCoefficient(zVecCoef, modeCoef)

                xComp.ProjectCoefficient(xCoef)
                yComp.ProjectCoefficient(yCoef)
                zComp.ProjectCoefficient(zCoef)

                max_x = GetScalarMax(xComp)
                max_y = GetScalarMax(yComp)
                max_z = GetScalarMax(zComp)
                max_r = np.max([max_x, max_y, max_z])

                x_cmd = make_cmd('X', 0, 0, max_r, i)
                y_cmd = make_cmd('Y', 403, 0, max_r, i)
                z_cmd = make_cmd('Z', 800, 0, max_r, i)

                send_data_to_glvis(mode_x_sock, x, x_cmd)
                send_data_to_glvis(mode_y_sock, y, y_cmd)
                send_data_to_glvis(mode_z_sock, z, z_cmd)

                dmodeCoeff = mfem.VectorGridFunctionCoefficient(dx)
                dyCoef = mfem.InnerProductCoefficient(yVecCoef, dmodeCoef)
                dzCoef = mfem.InnerProductCoefficient(zVecCoef, dmodeCoef)
                zCoef = mfem.InnerProductCoefficient(zVecCoef, modeCoef)

                dyComp.ProjectCoefficient(dyCoef)
                dzComp.ProjectCoefficient(dzCoef)

                min_d = max_r / bbMax[0] - bbMin[0]
                max_y = GetScalarMax(dyComp)
                max_z = GetScalarMax(dzComp)
                max_r = np.max([max_y, max_z, max_d])

                dy_cmd = make_cmd('Y', 403, 375, max_r, i,
                                  title="Curl Eigenmode")
                dz_cmd = make_cmd('Z', 800, 375, max_r, i,
                                  title="Curl Eigenmode")

                send_data_to_glvis(mode_dy_sock, dyComp, dy_cmd)
                send_data_to_glvis(mode_dz_sock, dzComp, dz_cmd)

                ask_exit()

        elif dim == 2:
            mode_xy_sock = make_socketstrema()
            mode_z_sock = make_socketstrema()
            mode_dxy_sock = make_socketstrema()
            mode_dz_sock = make_socketstrema()

            mat = array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
            xyMat = mfem.DenseMatrix(mat)
            xyMatCoef = mfem.MatrixConstantCoefficient(xyMat)
            zVec = array([0, 0, 1])
            zVecCoef = mfem.VectorConstantCoefficient(zVec)

            fec_h1 = mfem.H1_FECollection(order, dim)
            fec_nd_xy = mfem.ND_FECollection(order, dim)
            fec_rt_xy = mfem.RT_FECollection(order-1, dim)
            fec_l2 = mfem.L2_FECollection(order-1, dim)

            fes_h1 = mfem.ParFiniteElementSpace(pmesh, fec_h1)
            fes_nd = mfem.ParFiniteElementSpace(pmesh, fec_nd_xy)
            fes_rt = mfem.ParFiniteElementSpace(pmesh, fec_rt_xy)
            fes_l2 = mfem.ParFiniteElementSpace(pmesh, fec_l2)

            xyComp = mfem.ParGridFunction(fes_nd)
            zComp = mfem.ParGridFunction(fes_h1)

            dxyComp = mfem.ParGridFunction(fes_rt)
            dzComp = mfem.ParGridFunction(fes_l2)

            for i in range(nev):
                if (myid == 0):
                    print("Eigenmode " + str(i+1) + '/' + str(nev) +
                          ", Lambda = " + "{:g}".format(eigenvalues[i]))

                x.Assign(ame.GetEigenvector(i))
                curl.Mult(x, dx)

                modeCoef = mfem.VectorGridFunctionCoefficient(x)

                xyCoef = mfem.MatrixVectorProductCoefficient(
                    xyMatCoef, modeCoef)
                zCoef = mfem.InnerProductCoefficient(zVecCoef, modeCoef)

                xyComp.ProjectCoefficient(xyCoef)
                zComp.ProjectCoefficient(zCoef)

                max_v = GetVectorMax(2, xyComp)
                max_s = GetScalarMax(zComp)
                max_r = np.max([max_v, max_s])

                xy_cmd = make_cmd('XY', 0, 0, max_r, i)
                z_cmd = make_cmd('Z', 403, 0, max_r, i)

                send_data_to_glvis(mode_xy_sock, xyComp, xy_cmd)
                send_data_to_glvis(mode_z_sock, zComp, z_cmd)

                dmodeCoef = mfem.VectorGridFunctionCoefficient(dx)
                dxyCoef = mfem.MatrixVectorProductCoefficient(
                    xyMatCoef, dmodeCoef)
                dzCoef = mfem.InnerProductCoefficient(zVecCoef, dmodeCoef)

                dxyComp.ProjectCoefficient(dxyCoef)
                dzComp.ProjectCoefficient(dzCoef)

                min_d = max_r / \
                    min([bbMax[0] - bbMin[0],  bbMax[1] - bbMin[1]])
                max_v = GetVectorMax(2, dxyComp)
                max_s = GetScalarMax(dzComp)
                max_r = np.max([max_v, max_s])

                dxy_cmd = make_cmd('XY', 0, 375, max_r, i,
                                   title="Curl Eigenmode")
                dz_cmd = make_cmd('Z', 403, 375, max_r, i,
                                  title="Curl Eigenmode")

                send_data_to_glvis(mode_dxy_sock, dxyComp, dxy_cmd)
                send_data_to_glvis(mode_dz_sock, dzComp, dz_cmd)

                ask_exit()

        else:
            mode_sock = make_socketstrema()
            dmode_sock = make_socketstrema()

            for i in range(nev):
                if (myid == 0):
                    print("Eigenmode " + str(i+1) + '/' + str(nev) +
                          ", Lambda = " + "{:g}".format(eigenvalues[i]))

                x.Assign(ame.GetEigenvector(i))
                curl.Mult(x, dx)

                max_v = GetVectorMax(3, x)
                cmd = make_cmd('', 0, 0, max_v, i)
                max_v = GetVectorMax(3, dx)
                d_cmd = make_cmd('', 0, 375, max_v, i)

                send_data_to_glvis(mode_sock, x, cmd)
                send_data_to_glvis(dmode_sock, dx, d_cmd)

                ask_exit()


def GetVectorMax(vdim, x):
    zeroVec = mfem.Vector(vdim)
    zeroVec.Assign(0.0)
    zero = mfem.VectorConstantCoefficient(zeroVec)
    nrm = x.ComputeMaxError(zero)
    return nrm


def GetScalarMax(x):
    zero = mfem.ConstantCoefficient(0.0)
    nrm = x.ComputeMaxError(zero)
    return nrm


def make_socketstrema():
    sock = mfem.socketstream("localhost", 19916)
    sock.precision(8)
    return sock


if __name__ == "__main__":
    from mfem.common.arg_parser import ArgParser

    parser = ArgParser(description='Ex32p (Maxwell Eigenvalue Problem)')
    parser.add_argument('-m', '--mesh',
                        default="inline-quad.mesh",
                        action='store', type=str,
                        help='Mesh file to use.')
    parser.add_argument('-rs', '--refine-serial',
                        action='store', default=2, type=int,
                        help="Number of times to refine the mesh uniformly in serial")
    parser.add_argument('-rp', '--refine-parallel',
                        action='store', default=1, type=int,
                        help="Number of times to refine the mesh uniformly in paralle.")

    parser.add_argument('-o', '--order',
                        action='store', default=1, type=int,
                        help="Finite element order (polynomial degree) or -1 for isoparametric space.")
    parser.add_argument("-n", "--num-eigs",
                        action='store',
                        type=float,
                        default=5,
                        help="Number of eigen values to compute")
    parser.add_argument('-no-vis', '--no-visualization',
                        action='store_true',
                        help='Enable GLVis visualization')

    args = parser.parse_args()
    if myid == 0:
        parser.print_options(args)

    order = args.order
    meshfile = expanduser(
        join(os.path.dirname(__file__), '..', 'data', args.mesh))
    visualization = not args.no_visualization

    run(nev=args.num_eigs,
        order=order,
        rs=args.refine_serial,
        rp=args.refine_parallel,
        meshfile=meshfile,
        visualization=visualization)



'''
   MFEM example 10p

      This examples solves a time dependent nonlinear elasticity
      problem of the form dv/dt = H(x) + S v, dx/dt = v, where H is a
      hyperelastic model and S is a viscosity operator of Laplacian
      type.

   How to run:
      mpirun -np 4 python <arguments>

   Example of arguments:
      ex10p.py -m beam-quad.mesh -s 3 -rs 2 -dt 3
      ex10p.py -m beam-tri.mesh -s 3 -rs 2 -dt 3
      ex10p.py -m beam-hex.mesh -s 2 -rs 1 -dt 3
      ex10p.py -m beam-tet.mesh -s 2 -rs 1 -dt 3
      ex10p.py -m beam-quad.mesh -s 14 -rs 2 -dt 0.03 -vs 20
      ex10p.py -m beam-hex.mesh -s 14 -rs 1 -dt 0.05 -vs 20

   See c++ version in the MFEM library for more detail 
'''
import sys
import mfem.par as mfem

from mfem.common.arg_parser import ArgParser
from os.path import expanduser, join, dirname
import numpy as np
from numpy import sqrt, pi, cos, sin, hypot, arctan2
from scipy.special import erfc

from mfem.par import intArray, add_vector
from mpi4py import MPI

num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank

parser = ArgParser(description='Ex10p')
parser.add_argument('-m', '--mesh',
                    default='beam-quad.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-rs', '--refine-serial',
                    action='store', default=2, type=int,
                    help="Number of times to refine the mesh uniformly before parallel")
parser.add_argument('-rp', '--refine-parallel',
                    action='store', default=0, type=int,
                    help="Number of times to refine the mesh uniformly after parallel")
parser.add_argument('-o', '--order',
                    action='store', default=2, type=int,
                    help="Finite element order (polynomial degree)")
help_ode = "\n".join(["ODE solver: 1 - Backward Euler, 2 - SDIRK2, 3 - SDIRK3",
                      "\t11 - Forward Euler, 12 - RK2",
                      "\t13 - RK3 SSP, 14 - RK4."])
parser.add_argument('-s', '--ode-solver',
                    action='store', default=3, type=int,
                    help=help_ode)
parser.add_argument('-tf', '--t-final',
                    action='store', default=300.0, type=float,
                    help="Final time; start time is 0.")
parser.add_argument('-dt', '--time-step',
                    action='store', default=3.0, type=float,
                    help="Time step")

parser.add_argument("-v", "--viscosity",
                    action='store', default=1e-2, type=float,
                    help="Viscosity coefficient.")
parser.add_argument("-mu", "--shear-modulus",
                    action='store', default=0.25, type=float,
                    help="Shear modulus in the Neo-Hookean hyperelastic model.")
parser.add_argument("-K", "--bulk-modulus",
                    action='store', default=5.0, type=float,
                    help="Bulk modulus in the Neo-Hookean hyperelastic model.")
parser.add_argument('-vis', '--visualization',
                    action='store_true', default=True,
                    help='Enable GLVis visualization')
parser.add_argument("-vs", "--visualization-steps",
                    action='store', default=1,  type=int,
                    help="Visualize every n-th timestep.")
args = parser.parse_args()

ser_ref_levels = args.refine_serial
par_ref_levels = args.refine_parallel
order = args.order
ode_solver_type = args.ode_solver
t_final = args.t_final
dt = args.time_step
visc = args.viscosity
mu = args.shear_modulus
K = args.bulk_modulus
visualization = args.visualization
vis_steps = args.visualization_steps
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

# 3. Read the serial mesh from the given mesh file on all processors. We can
#    handle triangular, quadrilateral, tetrahedral and hexahedral meshes
#    with the same code.
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()

# 4. Define the ODE solver used for time integration. Several implicit
#    singly diagonal implicit Runge-Kutta (SDIRK) methods, as well as
#    explicit Runge-Kutta methods are available.
if ode_solver_type == 1:
    ode_solver = BackwardEulerSolver()
elif ode_solver_type == 2:
    ode_solver = mfem.SDIRK23Solver(2)
elif ode_solver_type == 3:
    ode_solver = mfem.SDIRK33Solver()
elif ode_solver_type == 11:
    ode_solver = ForwardEulerSolver()
elif ode_solver_type == 12:
    ode_solver = mfem.RK2Solver(0.5)
elif ode_solver_type == 13:
    ode_solver = mfem.RK3SSPSolver()
elif ode_solver_type == 14:
    ode_solver = mfem.RK4Solver()
elif ode_solver_type == 22:
    ode_solver = mfem.ImplicitMidpointSolver()
elif ode_solver_type == 23:
    ode_solver = mfem.SDIRK23Solver()
elif ode_solver_type == 24:
    ode_solver = mfem.SDIRK34Solver()
else:
    if myid == 0:
        print("Unknown ODE solver type: " + str(ode_solver_type))
    sys.exit()

# 5. Refine the mesh in serial to increase the resolution. In this example
#    we do 'ser_ref_levels' of uniform refinement, where 'ser_ref_levels' is
#    a command-line parameter.
for lev in range(ser_ref_levels):
    mesh.UniformRefinement()

# 6. Define a parallel mesh by a partitioning of the serial mesh. Refine
#    this mesh further in parallel to increase the resolution. Once the
#    parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
for x in range(par_ref_levels):
    pmesh.UniformRefinement()

# 7. Define the parallel vector finite element spaces representing the mesh
#    deformation x_gf, the velocity v_gf, and the initial configuration,
#    x_ref. Define also the elastic energy density, w_gf, which is in a
#    discontinuous higher-order space. Since x and v are integrated in time
#    as a system, we group them together in block vector vx, on the unique
#    parallel degrees of freedom, with offsets given by array true_offset.

fec = mfem.H1_FECollection(order, dim)
fespace = mfem.ParFiniteElementSpace(pmesh, fec, dim)
glob_size = fespace.GlobalTrueVSize()
if (myid == 0):
    print('Number of velocity/deformation unknowns: ' + str(glob_size))

true_size = fespace.TrueVSize()
true_offset = mfem.intArray(3)
true_offset[0] = 0
true_offset[1] = true_size
true_offset[2] = 2*true_size

vx = mfem.BlockVector(true_offset)

v_gf = mfem.ParGridFunction(fespace)
x_gf = mfem.ParGridFunction(fespace)

x_ref = mfem.ParGridFunction(fespace)
pmesh.GetNodes(x_ref)

w_fec = mfem.L2_FECollection(order + 1, dim)
w_fespace = mfem.ParFiniteElementSpace(pmesh, w_fec)
w_gf = mfem.ParGridFunction(w_fespace)

# 8. Set the initial conditions for v_gf, x_gf and vx, and define the
#    boundary conditions on a beam-like mesh (see description above).


class InitialVelocity(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        dim = len(x)
        s = 0.1/64.

        v = np.zeros(len(x))
        v[-1] = s*x[0]**2*(8.0-x[0])
        v[0] = -s*x[0]**2
        return v


class InitialDeformation(mfem.VectorPyCoefficient):
    def EvalValue(self, x):
        return x.copy()


velo = InitialVelocity(dim)
v_gf.ProjectCoefficient(velo)
deform = InitialDeformation(dim)
x_gf.ProjectCoefficient(deform)

v_gf.GetTrueDofs(vx.GetBlock(0))
x_gf.GetTrueDofs(vx.GetBlock(1))

ess_bdr = mfem.intArray(fespace.GetMesh().bdr_attributes.Max())
ess_bdr.Assign(0)
ess_bdr[0] = 1


# 9. Initialize the hyperelastic operator, the GLVis visualization and print
#    the initial energies.
class ElasticEnergyCoefficient(mfem.PyCoefficient):
    def __init__(self, model, x):
        self.x = x
        self.model = model
        self.J = mfem.DenseMatrix()
        mfem.PyCoefficient.__init__(self)

    def Eval(self, T, ip):
        self.model.SetTransformation(T)
        self.x.GetVectorGradient(T, self.J)
        return self.model.EvalW(self.J)/(self.J.Det())


class ReducedSystemOperator(mfem.PyOperator):
    def __init__(self, M, S, H, ess_tdof_list):
        mfem.PyOperator.__init__(self, M.ParFESpace().TrueVSize())
        self.M = M
        self.S = S
        self.H = H
        self.Jacobian = None
        h = M.ParFESpace().TrueVSize()
        self.w = mfem.Vector(h)
        self.z = mfem.Vector(h)
        self.dt = 0.0
        self.ess_tdof_list = ess_tdof_list

    def SetParameters(self, dt, v, x):
        self.dt = dt
        self.v = v
        self.x = x

    def Mult(self, k, y):
        add_vector(self.v, self.dt, k, self.w)
        add_vector(self.x, self.dt, self.w, self.z)
        self.H.Mult(self.z, y)
        self.M.TrueAddMult(k, y)
        self.S.TrueAddMult(self.w, y)
        y.SetSubVector(self.ess_tdof_list, 0.0)

    def GetGradient(self, k):
        localJ = mfem.Add(1.0, self.M.SpMat(), self.dt, self.S.SpMat())
        add_vector(self.v, self.dt, k, self.w)
        add_vector(self.x, self.dt, self.w, self.z)
        localJ.Add(self.dt * self.dt,  self.H.GetLocalGradient(self.z))
        Jacobian = self.M.ParallelAssemble(localJ)
        Jacobian.EliminateRowsCols(self.ess_tdof_list)
        return Jacobian


class HyperelasticOperator(mfem.PyTimeDependentOperator):
    def __init__(self, fespace, ess_bdr, visc, mu, K):
        mfem.PyTimeDependentOperator.__init__(self, 2*fespace.TrueVSize(), 0.0)

        rel_tol = 1e-8
        skip_zero_entries = 0
        ref_density = 1.0

        self.ess_tdof_list = intArray()
        self.z = mfem.Vector(self.Height()//2)
        self.fespace = fespace
        self.viscosity = visc
        self.newton_solver = mfem.NewtonSolver(fespace.GetComm())

        M = mfem.ParBilinearForm(fespace)
        S = mfem.ParBilinearForm(fespace)
        H = mfem.ParNonlinearForm(fespace)
        self.M = M
        self.H = H
        self.S = S

        rho = mfem.ConstantCoefficient(ref_density)
        M.AddDomainIntegrator(mfem.VectorMassIntegrator(rho))
        M.Assemble(skip_zero_entries)
        M.Finalize(skip_zero_entries)
        self.Mmat = M.ParallelAssemble()

        fespace.GetEssentialTrueDofs(ess_bdr, self.ess_tdof_list)
        self.Mmat.EliminateRowsCols(self.ess_tdof_list)

        M_solver = mfem.CGSolver(fespace.GetComm())
        M_prec = mfem.HypreSmoother()
        M_solver.iterative_mode = False
        M_solver.SetRelTol(rel_tol)
        M_solver.SetAbsTol(0.0)
        M_solver.SetMaxIter(30)
        M_solver.SetPrintLevel(0)
        M_prec.SetType(mfem.HypreSmoother.Jacobi)
        M_solver.SetPreconditioner(M_prec)
        M_solver.SetOperator(self.Mmat)

        self.M_solver = M_solver
        self.M_prec = M_prec

        model = mfem.NeoHookeanModel(mu, K)
        H.AddDomainIntegrator(mfem.HyperelasticNLFIntegrator(model))
        H.SetEssentialTrueDofs(self.ess_tdof_list)
        self.model = model

        visc_coeff = mfem.ConstantCoefficient(visc)
        S.AddDomainIntegrator(mfem.VectorDiffusionIntegrator(visc_coeff))
        S.Assemble(skip_zero_entries)
        S.Finalize(skip_zero_entries)

        self.reduced_oper = ReducedSystemOperator(M, S, H, self.ess_tdof_list)

        J_hypreSmoother = mfem.HypreSmoother()
        J_hypreSmoother.SetType(mfem.HypreSmoother.l1Jacobi)
        J_hypreSmoother.SetPositiveDiagonal(True)
        J_prec = J_hypreSmoother

        J_minres = mfem.MINRESSolver(fespace.GetComm())
        J_minres.SetRelTol(rel_tol)
        J_minres.SetAbsTol(0.0)
        J_minres.SetMaxIter(300)
        J_minres.SetPrintLevel(-1)
        J_minres.SetPreconditioner(J_prec)

        self.J_solver = J_minres
        self.J_prec = J_prec

        newton_solver = mfem.NewtonSolver(fespace.GetComm())
        newton_solver.iterative_mode = False
        newton_solver.SetSolver(self.J_solver)
        newton_solver.SetOperator(self.reduced_oper)
        newton_solver.SetPrintLevel(1)  # print Newton iterations
        newton_solver.SetRelTol(rel_tol)
        newton_solver.SetAbsTol(0.0)
        newton_solver.SetAdaptiveLinRtol(2, 0.5, 0.9)
        newton_solver.SetMaxIter(10)
        self.newton_solver = newton_solver

    def Mult(self, vx, vx_dt):
        sc = self.Height()//2
        v = mfem.Vector(vx, 0,  sc)
        x = mfem.Vector(vx, sc,  sc)
        dv_dt = mfem.Vector(dvx_dt, 0, sc)
        dx_dt = mfem.Vector(dvx_dt, sc,  sc)

        self.H.Mult(x, z)
        if (self.viscosity != 0.0):
            S.TrueAddMult(v, z)
            z.SetSubVector(self.ess_tdof_list, 0.0)
        z.Neg()
        self.M_solver.Mult(z, dv_dt)
        dx_dt = v

    def ImplicitSolve(self, dt, vx, dvx_dt):
        sc = self.Height()//2
        v = mfem.Vector(vx, 0,  sc)
        x = mfem.Vector(vx, sc,  sc)
        dv_dt = mfem.Vector(dvx_dt, 0, sc)
        dx_dt = mfem.Vector(dvx_dt, sc,  sc)

        # By eliminating kx from the coupled system:
        # kv = -M^{-1}*[H(x + dt*kx) + S*(v + dt*kv)]
        # kx = v + dt*kv
        # we reduce it to a nonlinear equation for kv, represented by the
        # backward_euler_oper. This equation is solved with the newton_solver
        # object (using J_solver and J_prec internally).
        self.reduced_oper.SetParameters(dt, v, x)
        zero = mfem.Vector()  # empty vector is interpreted as
        # zero r.h.s. by NewtonSolver
        self.newton_solver.Mult(zero, dv_dt)
        add_vector(v, dt, dv_dt, dx_dt)

    def ElasticEnergy(self, x):
        return self.H.GetEnergy(x)

    def KineticEnergy(self, v):
        local_energy = 0.5*self.M.InnerProduct(v, v)
        energy = MPI.COMM_WORLD.allreduce(local_energy, op=MPI.SUM)
        return energy

    def GetElasticEnergyDensity(self, x, w):
        w_coeff = ElasticEnergyCoefficient(self.model, x)
        w.ProjectCoefficient(w_coeff)


def visualize(out, pmesh, deformed_nodes, field,
              field_name='', init_vis=False):
    nodes = deformed_nodes
    owns_nodes = 0

    nodes, owns_nodes = pmesh.SwapNodes(nodes, owns_nodes)

    out.send_text("parallel " + str(num_procs) + " " + str(myid))
    out.send_solution(pmesh, field)

    nodes, owns_nodes = pmesh.SwapNodes(nodes, owns_nodes)

    if (init_vis):
        out.send_text("window_size 400 400")
        out.send_text("window_title '" + field_name)
        if (pmesh.SpaceDimension() == 2):
            out.send_text("view 0 0")
            out.send_text("keys jl")
        out.send_text("keys cm")         # show colorbar and mesh
        # update value-range; keep mesh-extents fixed
        out.send_text("autoscale value")
        out.send_text("pause")
    out.flush()


oper = HyperelasticOperator(fespace, ess_bdr, visc,  mu, K)
if (visualization):
    vis_v = mfem.socketstream("localhost", 19916)
    vis_v.precision(8)
    visualize(vis_v, pmesh, x_gf, v_gf, "Velocity", True)

    MPI.COMM_WORLD.Barrier()
    vis_w = mfem.socketstream("localhost", 19916)
    oper.GetElasticEnergyDensity(x_gf, w_gf)
    vis_w.precision(8)
    visualize(vis_w, pmesh, x_gf, w_gf, "Elastic energy density", True)

ee0 = oper.ElasticEnergy(x_gf)
ke0 = oper.KineticEnergy(v_gf)

if myid == 0:
    print("initial elastic energy (EE) = " + str(ee0))
    print("initial kinetic energy (KE) = " + str(ke0))
    print("initial   total energy (TE) = " + str(ee0 + ke0))

# 8. Perform time-integration (looping over the time iterations, ti, with a
#    time-step dt).
t = 0.
ti = 1

oper.SetTime(t)
ode_solver.Init(oper)
last_step = False

while not last_step:
    dt_real = min(dt, t_final - t)
    t, dt = ode_solver.Step(vx, t, dt_real)

    if (t >= t_final - 1e-8*dt):
        last_step = True

    if (last_step or (ti % vis_steps) == 0):
        v_gf.Distribute(vx.GetBlock(0))
        x_gf.Distribute(vx.GetBlock(1))

        ee = oper.ElasticEnergy(x_gf)
        ke = oper.KineticEnergy(v_gf)

        text = ("step " + str(ti) + ", t = " + str(t) +
                ", EE = " + "{:g}".format(ee) +
                ", KE = " + "{:g}".format(ke) +
                ", dTE = " + "{:g}".format((ee+ke)-(ee0+ke0)))

        if myid == 0:
            print(text)
        if visualization:
            visualize(vis_v, pmesh, x_gf, v_gf)
            oper.GetElasticEnergyDensity(x_gf, w_gf)
            visualize(vis_w, pmesh, x_gf, w_gf)

    ti = ti + 1

#
# if i translate c++ line-by-line, ti seems the second swap does not work...
#

smyid = '{:0>6d}'.format(myid)
mesh_name = "deformed."+smyid
velo_name = "velocity."+smyid
ee_name = "elastic_energy."+smyid

nodes = x_gf
owns_nodes = 0
nodes, owns_nodes = pmesh.SwapNodes(nodes, owns_nodes)
pmesh.Print(mesh_name, 8)
pmesh.SwapNodes(nodes, owns_nodes)

v_gf.Save(velo_name, 8)
oper.GetElasticEnergyDensity(x_gf, w_gf)
w_gf.Save(ee_name,  8)



'''
   MFEM example 2

   See c++ version in the MFEM library for more detail 

   How to run:
      mpirun -np 2 python2.7 <arguments>
  
   Example of arguments:
      ex2p.py -m star.mesh
      ex2p.py -m square-disc.mesh
      ex2p.py -m escher.mesh
      ex2p.py -m fichera.mesh
      ex2p.py -m beam-tri.mesh -o 2 -sys   
      ex2p.py -m beam-quad.mesh -o 3 -elast
      ex2p.py -m beam-quad.mesh -o 3 -sc
'''
import numpy as np
from os.path import expanduser, join, dirname
import sys
from mfem import path
from mfem.common.arg_parser import ArgParser
import mfem.par as mfem
from mpi4py import MPI
num_procs = MPI.COMM_WORLD.size
myid = MPI.COMM_WORLD.rank


parser = ArgParser(description='Ex2 (linear elastisity)')
parser.add_argument('-m', '--mesh',
                    default='beam-tri.mesh',
                    action='store', type=str,
                    help='Mesh file to use.')
parser.add_argument('-vis', '--visualization',
                    action='store_true',
                    help='Enable GLVis visualization')
parser.add_argument('-o', '--order',
                    action='store', default=1, type=int,
                    help="Finite element order (polynomial degree) or -1 for isoparametric space.")
parser.add_argument('-sc', '--static-condensation',
                    action='store_false',
                    help="Enable static condensation.")
parser.add_argument('-elast', '--amg-for-elasticity',
                    action='store_true',
                    help='Use the special AMG elasticity solver (GM/LN approaches)',
                    dest='amg_elast', default=False)
parser.add_argument('-sys', '--amg-for-systems',
                    action='store_false',
                    help='Use  standard AMG for systems (unknown approach).',
                    dest='amg_elast', default=True)
args = parser.parse_args()

order = args.order
static_cond = args.static_condensation
amg_elast = args.amg_elast
visualization = args.visualization
if (myid == 0):
    parser.print_options(args)

device = mfem.Device('cpu')
if myid == 0:
    device.Print()

#  3. Read the (serial) mesh from the given mesh file on all processors.  We
#     can handle triangular, quadrilateral, tetrahedral, hexahedral, surface
#     and volume meshes with the same code.
meshfile = expanduser(join(dirname(__file__), '..', 'data', args.mesh))
mesh = mfem.Mesh(meshfile, 1, 1)
dim = mesh.Dimension()
if (mesh.attributes.Max() < 2 or mesh.bdr_attributes.Max() < 2):
    if (myid == 0):
        print('\n'.join(['Input mesh should have at least two materials and',
                         'two boundary attributes! (See schematic in ex2.cpp)']))
    sys.exit()

#  4. Select the order of the finite element discretization space. For NURBS
#     meshes, we increase the order by degree elevation.
if (mesh.NURBSext and order > mesh.NURBSext.GetOrder()):
    mesh.DegreeElevate(order - mesh.NURBSext.GetOrder())

#  5. Refine the serial mesh on all processors to increase the resolution. In
#     this example we do 'ref_levels' of uniform refinement. We choose
#     'ref_levels' to be the largest number that gives a final mesh with no
#     more than 1,000 elements.
ref_levels = int(np.floor(np.log(1000./mesh.GetNE())/np.log(2.)/dim))
for x in range(ref_levels):
    mesh.UniformRefinement()

#  6. Define a parallel mesh by a partitioning of the serial mesh. Refine
#     this mesh further in parallel to increase the resolution. Once the
#     parallel mesh is defined, the serial mesh can be deleted.
pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
del mesh
par_ref_levels = 1
for l in range(par_ref_levels):
    pmesh.UniformRefinement()

#   7. Define a parallel finite element space on the parallel mesh. Here we
#      use vector finite elements, i.e. dim copies of a scalar finite element
#      space. We use the ordering by vector dimension (the last argument of
#      the FiniteElementSpace constructor) which is expected in the systems
#      version of BoomerAMG preconditioner. For NURBS meshes, we use the
#      (degree elevated) NURBS space associated with the mesh nodes.
use_nodal_fespace = pmesh.NURBSext and not amg_elast
if use_nodal_fespace:
    fespace = pmesh.GetNodes().FESpace()
else:
    fec = mfem.H1_FECollection(order, dim)
    fespace = mfem.ParFiniteElementSpace(pmesh, fec, dim,
                                         mfem.Ordering.byVDIM)

size = fespace.GlobalTrueVSize()
if (myid == 0):
    print("Number of finite element unknowns: " + str(size))

# 8. Determine the list of true (i.e. parallel conforming) essential
#    boundary dofs. In this example, the boundary conditions are defined by
#    marking only boundary attribute 1 from the mesh as essential and
#    converting it to a list of true dofs.
ess_bdr = mfem.intArray(pmesh.bdr_attributes.Max())
ess_tdof_list = mfem.intArray()
ess_bdr.Assign(0)
ess_bdr[0] = 1
fespace.GetEssentialTrueDofs(ess_bdr, ess_tdof_list)
print("here")

#  9. Set up the parallel linear form b(.) which corresponds to the
#     right-hand side of the FEM linear system. In this case, b_i equals the
#     boundary integral of f*phi_i where f represents a "pull down" force on
#     the Neumann part of the boundary and phi_i are the basis functions in
#     the finite element fespace. The force is defined by the object f, which
#     is a vector of Coefficient objects. The fact that f is non-zero on
#     boundary attribute 2 is indicated by the use of piece-wise constants
#     coefficient for its last component.
f = mfem.VectorArrayCoefficient(dim)
for i in range(dim-1):
    f.Set(i, mfem.ConstantCoefficient(0.0))

pull_force = mfem.Vector([0]*pmesh.bdr_attributes.Max())
pull_force[1] = -1.0e-2
f.Set(dim-1, mfem.PWConstCoefficient(pull_force))


b = mfem.ParLinearForm(fespace)
b.AddBoundaryIntegrator(mfem.VectorBoundaryLFIntegrator(f))
print('r.h.s. ...')
b.Assemble()

# 10. Define the solution vector x as a parallel finite element grid
#     function corresponding to fespace. Initialize x with initial guess of
#     zero, which satisfies the boundary conditions.
x = mfem.GridFunction(fespace)
x.Assign(0.0)
print('here')
# 11. Set up the parallel bilinear form a(.,.) on the finite element space
#     corresponding to the linear elasticity integrator with piece-wise
#     constants coefficient lambda and mu.

lamb = mfem.Vector(pmesh.attributes.Max())
lamb.Assign(1.0)
lamb[0] = lamb[1]*50
lambda_func = mfem.PWConstCoefficient(lamb)
mu = mfem.Vector(pmesh.attributes.Max())
mu.Assign(1.0)
mu[0] = mu[1]*50
mu_func = mfem.PWConstCoefficient(mu)
a = mfem.ParBilinearForm(fespace)
a.AddDomainIntegrator(mfem.ElasticityIntegrator(lambda_func, mu_func))

#  12. Assemble the parallel bilinear form and the corresponding linear
#      system, applying any necessary transformations such as: parallel
#      assembly, eliminating boundary conditions, applying conforming
#      constraints for non-conforming AMR, static condensation, etc.
if (myid == 0):
    print('matrix...')
if (static_cond):
    a.EnableStaticCondensation()
a.Assemble()

A = mfem.HypreParMatrix()
B = mfem.Vector()
X = mfem.Vector()
a.FormLinearSystem(ess_tdof_list, x, b, A, X, B)
if (myid == 0):
    print('...done')
    print("Size of linear system: " + str(A.GetGlobalNumRows()))

#  13. Define and apply a parallel PCG solver for A X = B with the BoomerAMG
#      preconditioner from hypre.
amg = mfem.HypreBoomerAMG(A)
if (amg_elast and not a.StaticCondensationIsEnabled()):
    amg.SetElasticityOptions(fespace)
else:
    amg.SetSystemsOptions(dim)
pcg = mfem.HyprePCG(A)
pcg.SetTol(1e-8)
pcg.SetMaxIter(500)
pcg.SetPrintLevel(2)
pcg.SetPreconditioner(amg)
pcg.Mult(B, X)

#  14. Recover the parallel grid function corresponding to X. This is the
#      local finite element solution on each processor.
a.RecoverFEMSolution(X, b, x)

#  15. For non-NURBS meshes, make the mesh curved based on the finite element
#      space. This means that we define the mesh elements through a fespace
#      based transformation of the reference element.  This allows us to save
#      the displaced mesh as a curved mesh when using high-order finite
#      element displacement field. We assume that the initial mesh (read from
#      the file) is not higher order curved mesh compared to the chosen FE
#      space.
if (not use_nodal_fespace):
    pmesh.SetNodalFESpace(fespace)

#  16. Save in parallel the displaced mesh and the inverted solution (which
#      gives the backward displacements to the original grid). This output
#      can be viewed later using GLVis: "glvis -np <np> -m mesh -g sol".

nodes = pmesh.GetNodes()
nodes += x
x *= -1

smyid = '{:0>6d}'.format(myid)
mesh_name = "mesh."+smyid
sol_name = "sol."+smyid

pmesh.Print(mesh_name, 8)
x.Save(sol_name, 8)

#  17. Send the above data by socket to a GLVis server.  Use the "n" and "b"
#      keys in GLVis to visualize the displacements.

if (visualization):
    sol_sock = mfem.socketstream("localhost", 19916)
    sol_sock.send_text("parallel " + str(num_procs) + " " + str(myid))
    sol_sock.precision(8)
    sol_sock.send_solution(pmesh,  x)



import os
import sys
import subprocess as sp
import time
import shutil
import re

skip_test = [''] 
#skip_test = ['ex15p']  #skip this since it takes too long on my Mac..;D

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

import resource

def format_memory_usage(point="memory usage"):
    usage=resource.getrusage(resource.RUSAGE_SELF)
    txt = "%s: usertime=%s systime=%s mem="+ bcolors.FAIL + "%s" + bcolors.ENDC + " mb"
    return txt%(point,usage[0],usage[1],
               (usage[2]*resource.getpagesize())/1000000.0 )


if __name__ == '__main__':
   import ex1p

   for i in range(30):
       ex1p.run()
       print(format_memory_usage())



'''
  Ex33_common.py

  This is a translation of MFEM ex33.hpp. LAPACK call, mfem.Vector, 
  mfem.DenseMatrix are replaced using numpy
 

                (Implementation of the AAA algorithm)

  Here, we implement the triple-A algorithm [1] for the rational approximation
  of complex-valued functions,

          p(z)/q(z) ≈ f(z).

  In this file, we always assume f(z) = z^{-α}. The triple-A algorithm
  provides a robust, accurate approximation in rational barycentric form.
  This representation must be transformed into a partial fraction
  representation in order to be used to solve a spectral FPDE.

  More specifically, we first expand the numerator in terms of the zeros of
  the rational approximation,

          p(z) ∝ Π_i (z - z_i),

  and expand the denominator in terms of the poles of the rational
  approximation,

          q(z) ∝ Π_i (z - p_i).

  We then use these zeros and poles to derive the partial fraction expansion

          f(z) ≈ p(z)/q(z) = Σ_i c_i / (z - p_i).

  [1] Nakatsukasa, Y., Sète, O., & Trefethen, L. N. (2018). The AAA algorithm
      for rational approximation. SIAM Journal on Scientific Computing, 40(3),
      A1494-A1522.
'''
import numpy as np
import scipy
from scipy.linalg import eig

import mfem
if mfem.mfem_mode == 'parallel':
    import mfem.par as mfem
    from mfem.par import intArray, doubleArray
else:
    import mfem.ser as mfem
    from mfem.ser import intArray, doubleArray

from sys import float_info
eps = float_info.min


def RationalApproximation_AAA(val, pt, tol, max_order):
    '''
    RationalApproximation_AAA: compute the rational approximation (RA) of data
    val at the set of points pt

    in:
        val        Vector of data values
        pt         Vector of sample points
        tol        Relative tolerance
        max_order  Maximum number of terms (order) of the RA
    out:
        z          Support points of the RA in rational barycentric form
        f          Data values at support points at z
        w          Weights of the RA in rational barycentric form

    See pg. A1501 of Nakatsukasa et al. [1].
    '''

    # number of sample points
    size = len(val)  # .Size()
    assert len(pt) == size, "size mismatch"

    # Initializations
    J = list(range(size))

    c_i = []
    # mean of the value vector
    mean_val = np.mean(val)
    R = np.array([mean_val]*size)

    z = []
    f = []
    w = []

    for k in range(max_order):
        # select next support point
        idx = 0
        tmp_max = 0

        idx = np.argmax(np.abs(val-R))
        # Append support points and data values
        z.append(pt[idx])
        f.append(val[idx])

        # Update index vector
        J.remove(idx)

        # next column in Cauchy matrix
        C_tmp = [(1.0/(pp-pt[idx]) if pp != pt[idx] else np.inf) for pp in pt]

        c_i = np.hstack((c_i, C_tmp))

        h_C = len(C_tmp)
        w_C = k+1

        # note: tranpose is necessary due to the difference of column-major
        # and raw-major of matrix
        C = c_i.reshape(w_C, h_C).transpose()
        Ctemp = C.copy()
        Ctemp = Ctemp*(np.atleast_2d(1/val)).transpose()  # InvLeftScaling
        Ctemp = Ctemp*f                                   # RgithScaling
        A = C - Ctemp
        A = A*(np.atleast_2d(val)).transpose()            # LeftScaling

        h_Am = len(J)
        w_Am = A.shape[1]
        Am = np.zeros((h_Am, w_Am))

        for i in range(h_Am):
            ii = J[i]
            for j in range(w_Am):
                Am[i, j] = A[ii, j]

        u, s, vh = np.linalg.svd(Am)

        w = vh[k, :]

        N = C.dot(w*np.array(f))
        D = C.dot(w)
        R = val.copy()
        for i, ii in enumerate(J):
            R[ii] = N[ii]/D[ii]
        verr = val - R

        if np.max(verr) <= tol*max(val):
            break

    return z, f, w


def ComputePolesAndZeros(z, f, w):
    '''
    ComputePolesAndZeros: compute the poles  and zeros of the
    rational function f(z) = C p(z)/q(z) from its ration barycentric form.

    in:
        z      Support points in rational barycentric form
        f      Data values at support points @a z
        w      Weights in rational barycentric form
    out:
        poles  Array of poles (roots of p(z))
        zeros  Array of zeros (roots of q(z))
        scale  Scaling constant in f(z) = C p(z)/q(z)

    See pg. A1501 of Nakatsukasa et al. [1].
    '''

    # Initialization
    poles = []
    zeros = []

    # Compute the poles
    m = len(w)
    B = np.zeros((m+1, m+1))
    E = np.zeros((m+1, m+1))

    for i in range(m+1):
        if i == 0:
            continue
        B[i, i] = 1.
        E[0, i] = w[i-1]
        E[i, 0] = 1.
        E[i, i] = z[i-1]

    # real part of eigen value
    evalues = eig(E, B, left=False, right=False).real
    new_poles = evalues[np.isfinite(evalues)]

    poles.extend(new_poles)

    B = np.zeros((m+1, m+1))
    E = np.zeros((m+1, m+1))
    for i in range(m+1):
        if i == 0:
            continue
        B[i, i] = 1.
        E[0, i] = w[i-1] * f[i-1]
        E[i, 0] = 1.
        E[i, i] = z[i-1]

    # real part of eigen value
    evalues = eig(E, B, left=False, right=False).real
    new_zeros = evalues[np.isfinite(evalues)]

    zeros.extend(new_zeros)

    scale = np.dot(w, f)/np.sum(w)

    return poles, zeros, scale


def PartialFractionExpansion(scale, poles, zeros):
    '''
    PartialFractionExpansion: compute the partial fraction expansion of the
    rational function f(z) = Σ_i c_i / (z - p_i) from its poles and zeros
    @a zeros [in].

    in: 
        poles   Array of poles (same as p_i above)
        zeros   Array of zeros
        scale   Scaling constant
    out:
        coeffs  Coefficients c_i 
    '''

    # Note: C p(z)/q(z) = Σ_i c_i / (z - p_i) results in an system of equations
    # where the N unknowns are the coefficients c_i. After multiplying the
    # system with q(z), the coefficients c_i can be computed analytically by
    # choosing N values for z. Choosing z_j = = p_j diagonalizes the system and
    # one can obtain an analytic form for the c_i coefficients. The result is
    # implemented in the code block below.

    psize = len(poles)
    zsize = len(zeros)
    coeffs = [scale] * psize

    for i in range(psize):
        tmp_numer = 1.0
        for j in range(zsize):
            tmp_numer *= poles[i]-zeros[j]

        tmp_denom = 1.0
        for k in range(psize):
            if k != i:
                tmp_denom *= poles[i]-poles[k]

        coeffs[i] *= tmp_numer / tmp_denom
    return coeffs


def ComputePartialFractionApproximation(alpha,
                                        lmax=1000.,
                                        tol=1e-10,
                                        npoints=1000,
                                        max_order=100):
    '''
    ComputePartialFractionApproximation: compute a rational approximation (RA)
    in partial fraction form, e.g., f(z) ≈ Σ_i c_i / (z - p_i), from sampled
    values of the function f(z) = z^{-a}, 0 < a < 1.

    in:
       alpha         Exponent a in f(z) = z^-a
       lmax,npoints  f(z) is uniformly sampled @a npoints times in the
                     interval [ 0, @a lmax ]
       tol           Relative tolerance
       max_order     Maximum number of terms (order) of the RA

    out:
       coeffs        Coefficients c_i
       poles         Poles p_i
    '''

    assert alpha < 1., "alpha must be less than 1"
    assert alpha > 0., "alpha must be greater than 0"
    assert npoints > 2, "npoints must be greater than 2"
    assert lmax > 0,  "lmin must be greater than 0"
    assert tol > 0,  "tol must be greater than 0"

    dx = lmax / (npoints-1)

    x = np.arange(npoints)*dx
    val = x**(1-alpha)

    # Apply triple-A algorithm to f(x) = x^{1-a}
    z, f, w = RationalApproximation_AAA(val,  # mfem.Vector(val),
                                        x,  # mfem.Vector(x),
                                        tol, max_order)

    # Compute poles and zeros for RA of f(x) = x^{1-a}
    poles, zeros, scale = ComputePolesAndZeros(z, f, w)

    # Remove the zero at x=0, thus, delivering a RA for f(x) = x^{-a}
    zeros.remove(0.0)

    # Compute partial fraction approximation of f(x) = x^{-a}
    coeffs = PartialFractionExpansion(scale, poles, zeros)

    return poles, coeffs



from __future__ import print_function
import os
import sys
import numpy as np


def run_test1(mfem):
    print("Test mesh module")
    Nvert = 6
    Nelem = 8
    Nbelem = 2

    mesh = mfem.Mesh(2, Nvert, Nelem, 2, 3)
    tri_v = [[1.,  0.,  0.], [0.,  1.,  0.], [-1.,  0.,  0.],
             [0., -1.,  0.], [0.,  0.,  1.], [0.,  0., -1.]]
    tri_e = [[0, 1, 4], [1, 2, 4], [2, 3, 4], [3, 0, 4],
             [1, 0, 5], [2, 1, 5], [3, 2, 5], [0, 3, 5]]
    tri_l = [[1, 4], [1, 2]]

    for j in range(Nvert):
        mesh.AddVertex(tri_v[j])
    for j in range(Nelem):
        mesh.AddTriangle(tri_e[j], 1)
    for j in range(Nbelem):
        mesh.AddBdrSegment(tri_l[j], 1)

    mesh.FinalizeTriMesh(1, 1, True)

    print(mesh.GetEdgeVertices(1))
    print(mesh.GetFaceElements(1))

    assert mesh.GetBdrElementAdjacentElement(1) == (
        1, 0), "GetBdrElementadjacentElement failed"

    i = 0
    Tr = mesh.GetElementTransformation(i)
    print(Tr.Weight())
    Tr = mesh.GetBdrElementTransformation(i)
    print(Tr.Weight())
    Tr = mesh.GetFaceTransformation(i)
    print(Tr.Weight())
    Tr = mesh.GetEdgeTransformation(i)
    print(Tr.Weight())


def run_test2(mfem):
    import tracemalloc
    import linecache
    mesh = mfem.Mesh(6, 6, 6, "TETRAHEDRON")

    r = mesh.CartesianPartitioning([2, 2, 2], return_list=True)
    print(r)

    r = mesh.CartesianPartitioning([2, 2, 2])
    print(r)
    mesh.PrintWithPartitioning(r, "partiotioned.mesh")
    '''
    tracemalloc.start()
    def display_top(snapshot, key_type='lineno', limit=3):
        snapshot = snapshot.filter_traces((
            tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
            tracemalloc.Filter(False, "<unknown>"),
        ))
        top_stats = snapshot.statistics(key_type)

        print("Top %s lines" % limit)
        for index, stat in enumerate(top_stats[:limit], 1):
            frame = stat.traceback[0]
            # replace "/path/to/module/file.py" with "module/file.py"
            filename = os.sep.join(frame.filename.split(os.sep)[-2:])
            print("#%s: %s:%s: %.1f KiB"
                  % (index, filename, frame.lineno, stat.size / 1024))
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                print('    %s' % line)

        other = top_stats[limit:]
        if other:
            size = sum(stat.size for stat in other)
            print("%s other: %.1f KiB" % (len(other), size / 1024))
        total = sum(stat.size for stat in top_stats)
        print("Total allocated size: %.1f KiB" % (total / 1024))

    snapshot = tracemalloc.take_snapshot()
    display_top(snapshot)
    '''
    '''
    d = mfem.intArray([2, 2, 2])
    dd = d.GetData()

    r = mesh.CartesianPartitioning(dd)
    result = mfem.intArray()
    result.MakeRef(r, mesh.GetNE())
    result.MakeDataOwner()
    '''

    '''    

    mesh.Print("sample.mesh")
    r = mesh.CartesianPartitioning([2, 2, 2])
    print(r)
    '''


def run_test3(mfem):
    m1 = mfem.Mesh.MakeCartesian3D(3, 3, 3, mfem.Element.TETRAHEDRON)
    m2 = mfem.Mesh.MakeCartesian3D(3, 3, 3, mfem.Element.TETRAHEDRON)

    for i in range(m2.GetNV()):
        vv = m2.GetVertexArray(i)
        vv += 2.5
    print(np.vstack(m1.GetVertexArray()))
    print(np.vstack(m2.GetVertexArray()))

    print("test if it fails (1)")
    try:
        m3 = mfem.Mesh.MakeMerged("error")
    except:
        import traceback
        print("exception happens")
        traceback.print_exc()
    print("test if it fails (2)")
    try:
        m3 = mfem.Mesh.MakeMerged(('hoge', m2))
    except:
        import traceback
        print("exception happens")
        traceback.print_exc()

    print("test if this one is ok (3)")
    m3 = mfem.Mesh.MakeMerged((m1, m2))
    assert m3.GetNV() == m1.GetNV() + m2.GetNV(), "merge did not work"
    print("test (3) ok")
    # m3.Print('merged_mesh2.mesh')


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '-p':
        import mfem.par as mfem
    else:
        import mfem.ser as mfem

    run_test1(mfem)
    run_test2(mfem)
    run_test3(mfem)



from __future__ import print_function
import os
from os.path import expanduser, join
import sys
import numpy as np
import io

from mfem import path as mfem_path

if len(sys.argv) > 1 and sys.argv[1] == '-p':
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid = MPI.COMM_WORLD.rank
    from mfem.common.parcsr_extra import ToScipyCSR, ToScipyCoo, ReadHypreDiag
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0


def get_inv_doftrans(doftrans):
    if doftrans is not None:
        Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
        vv = mfem.Vector(doftrans.Size())
        for i in range(doftrans.Size()):
            vv.Assign(0)
            vv[i] = 1
            doftrans.InvTransformPrimal(vv)
            Mdoftrans[:, i] = vv.GetDataArray()

        return Mdoftrans
    else:
        return None


def run_test():
    order = 1
    vdim = 1
    path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    meshfile = os.path.join(path, 'data', 'waveguide.mesh')

    mesh = mfem.Mesh(meshfile, 1, 1)

    dim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    fec = mfem.ND_FECollection(order, sdim)
    #fec = mfem.H1_FECollection(order, sdim)

    pmesh = mfem.ParMesh(MPI.COMM_WORLD, mesh)
    # pmesh.ReorientTetMesh()
    pfes = mfem.ParFiniteElementSpace(pmesh, fec, vdim)
    print('Number of global finite element unknowns: ' +
          str(pfes.GetTrueVSize()))

    P = pfes.Dof_TrueDof_Matrix()

    print("Is P HypreGPU?s:", mfem.is_HYPRE_USING_CUDA())

    print("Height/Width", P.Height(), P.Width())

    cparts = P.GetColPartArray()
    rparts = P.GetRowPartArray()
    m, n = P.M(), P.N()

    print("col part", cparts)
    print("row part", rparts)

    # this does not produce correct diag for non-square matrix !?
    #diag = mfem.Vector()
    # P.GetDiag(diag)
    #d1 = diag.GetDataArray()

    #d2 = ReadHypreDiag(P)
    #print("diag length:", d2.shape[0])

    merged = mfem.SparseMatrix()
    P.MergeDiagAndOffd(merged)

    print("checking number of rows: ", merged.Height())
    assert merged.Height() == P.Height(), "local height does not match"
    print("passed")
    #print("I", merged.GetIArray())
    #print("J", merged.GetJArray())

    print("checking length of local nnz: ", P.get_local_nnz())
    assert len(merged.GetDataArray()
               ) == P.get_local_nnz(), "local nnz not correct"
    print("passed")

    print("checking NNZ: ", P.NNZ())
    assert np.sum(MPI.COMM_WORLD.allgather(P.get_local_nnz())
                  ) == P.NNZ(), "total NNZ does not match"
    print("passed")

    from mfem.common.sparse_utils import sparsemat_to_scipycsr

    merged_csr = sparsemat_to_scipycsr(merged, float)

    # if mfem.is_HYPRE_USING_CUDA():
    Pcsr = ToScipyCSR(P)
    Pcsr2 = ToScipyCoo(P).tocsr()

    print("checking various matrix conversion")
    assert (merged_csr != Pcsr).nnz == 0, "ToScipyCSR deos not agree with merged matrix"
    assert (merged_csr !=
            Pcsr2).nnz == 0, "ToScipyCoo deos not agree with merged matrix"
    print("passed")

    print("size", Pcsr.indptr.shape, Pcsr.shape)
    VDoFtoGTDoF1 = Pcsr.indices

    print("Global V and True sizes", pfes.GlobalVSize(), pfes.GlobalTrueVSize())

    # print(pfes.GetNE())
    # for i in range(pfes.GetNE()):
    for i in [1]:
        vdofs = pfes.GetElementDofs(i)
        doftrans = pfes.GetElementDofTransformation(i)
        vdofs = [-1 - x if x < 0 else x for x in vdofs]
        ltdof = [pfes.GetLocalTDofNumber(ii) for ii in vdofs]
        gtdof1 = [list(Pcsr.indices[Pcsr.indptr[ii:ii+1]])
                  for ii in vdofs]  # if pfes.GetLocalTDofNumber(ii) >= 0]
        gtdof1 = sum(gtdof1, [])
        # if pfes.GetLocalTDofNumber(ii) >= 0]
        gtdof2 = [pfes.GetGlobalTDofNumber(ii) for ii in vdofs]
        print(gtdof1, gtdof2)
        print(gtdof1 != gtdof2)
        if True:
            print("myid: ", myid)
            # print(get_inv_doftrans(doftrans))
            print("Local DoFs", vdofs)
            print("Local TDoFs", ltdof)
            print("Global TDoFs from P             :", gtdof1)
            print("Global TDoFs GetGlobalTDoFNumber:", gtdof2)


if __name__ == "__main__":

    if not use_parallel:
        print("this is parallel only, skipping")
    else:
        run_test()



from __future__ import print_function
import os
import sys
import numpy as np

if len(sys.argv) > 1 and sys.argv[1] == '-p':   
    import mfem.par as mfem
    use_parallel = True
    from mfem.common.mpi_debug import nicePrint as print
    from mpi4py import MPI
    myid  = MPI.COMM_WORLD.rank
    
else:
    import mfem.ser as mfem
    use_parallel = False
    myid = 0
    
def run_test(): 
    m = mfem.DenseMatrix(3,3)
    m.Assign(np.arange(9.).reshape(3,3))
    m.Print()
    print(np.arange(9.).reshape(3,3))

    m = mfem.DenseMatrix(3,5)
    m.Assign(np.arange(15.).reshape(3,5))
    m.Print()
    print(np.arange(15.).reshape(3,5))
    
    
    x = np.zeros(5)+1
    y = np.zeros(5)+2
    z = np.zeros(5)+3
    mm = mfem.DenseMatrix(3, 5)
    mm.Assign(np.vstack([x, y, z]))
    mm.Print()
    mfem.Vector(mm.GetData(), 15).Print()

    mm.Print("densmat.dat")

    ### added this check based on the issue 99.
    import tracemalloc
    import gc
    tracemalloc.start()
    n = 4 * 1024 // 8
    d = 1024
    K = mfem.DenseMatrix(d,d)
    for i in range(10): 
        K.Assign(np.diag(d * np.ones(d)))
        data = K.GetDataArray()
        size, peak = tracemalloc.get_traced_memory()        
        #print(f"{size=}, {peak=}")

    m = mfem.DenseTensor(2, 3, 5)
    m.Assign(0)
    m[1, 2, 3] = 5
    print(m[1, 2, 3])
    
    assert np.all(m[3].GetDataArray() == m.GetDataArray()[3]), "First column slicing does not agree"

    t = np.arange(30.).reshape(2, 3, 5)
    size, peak = tracemalloc.get_traced_memory()        
    print(f"{size=}, {peak=}")
    
    m.Assign(t)
    data = []
    for i in range(m.SizeI()):
       for j in range(m.SizeJ()):
        for k in range(m.SizeK()):
            assert m[i, j, k] == t[i, j, k], "Tensor assigne fails"

    # these two should not agree (internal memory layout differs)        
    print(t.flatten())
    print(mfem.Vector(m.Data(), 30).GetDataArray())
    
if __name__=='__main__':
    run_test()
 
